/*
 *	container.h
 *
 *	This file defines the base classes of the OPaC container class hierarchy.
 *
 *	(C) Copyright 1994-1998, Pierre ARNAUD, OPaC bright ideas
 *		CH-1437 SUSCEVAZ, Switzerland. All rights reserved.
 */

/******************************************************************************

  This file is part of the OPaC class library ("OPaC").

  OPaC is distributed with NO WARRANTY OF ANY KIND.  No author or distributor
  accepts any responsibility for the consequences of using it, or for whether
  it serves any particular purpose or works at all, unless he or she says so
  in writing. Refer to the OPaC Free Public License (the "License") for full
  details.

  Every copy of OPaC must include a copy of the License, normally in a plain
  ASCII text file named PUBLIC. The License grants you the right to copy,
  modify and redistribute OPaC, but only under certain conditions described
  in the License. The copyright notice and this notice must be preserved on
  all copies.

******************************************************************************/

#ifndef	_OPaC_CONTAINER_H_
#define	_OPaC_CONTAINER_H_

#ifndef _OPaC_OPEXPORT_H_
#include <OPaC/opexport.h>
#endif

OP_INTERFACE (container.h);

/****************************************************************/

/*
 *	Very generic data container. It is a base class for any
 *	container, set, data base, etc. This one just specifies
 *	the access methods, not the modification methods.
 */

class OP_PUBLIC OPGenericContainer : public OPObject
{
	friend class OPIterator;
	
protected:
	OPListRef			iterators;				//	active iterators
	
protected:
	OPGenericContainer ();
	virtual ~ OPGenericContainer ();

	virtual void StoreInstance (OPStorageIn storage);	/* OVERRIDDEN */
	virtual void RestoreInstance (OPStorageIn storage);	/* OVERRIDDEN */
	
	virtual void InsertIterator (OPIteratorIn iterator);
	virtual void RemoveIterator (OPIteratorIn iterator);
	
public:
	virtual void Delete ();								/* OVERRIDDEN */
	
	virtual Count TotalCount () const = 0;
	virtual Bool Get (Count index, OPObjectOut object) = 0;
	virtual Bool SpecifyExtraction (OPArrayIn x) = 0;
	virtual OPIteratorRef NewIterator ();
	
	OP_DEFINE_CLASS (OPGenericContainer, OPObject);
};

/****************************************************************/

/*
 *	Less generic data container. It is used as a base class
 *	for the List, HashTable, DataBase and many more classes.
 */

class OP_PUBLIC OPContainer : public OPGenericContainer
{
	
protected:
	Count				num_element;			//	# elements in container
	
protected:
	OPContainer ();
	virtual ~ OPContainer ();
	
	virtual void StoreInstance (OPStorageIn storage);	/* OVERRIDDEN */
	virtual void RestoreInstance (OPStorageIn storage);	/* OVERRIDDEN */
	
public:
	virtual void Delete ();								/* OVERRIDDEN */
	
	virtual Bool Insert (OPObjectIn data) = 0;
	virtual Bool Remove (OPObjectIn data, Count nth = 0) = 0;
	virtual Bool Remove (Count index) = 0;
	virtual void RemoveAll () = 0;
	
	virtual Count TotalCount () const;					/* OVERRIDDEN */
	virtual Bool SpecifyExtraction (OPArrayIn x);		/* OVERRIDDEN */
	
	virtual Bool Get (Count index, OPObjectOut object) = 0;
	virtual Bool GetSimilar (OPObjectIn model, OPObjectOut object, Count nth = 0) = 0;
	
	virtual Bool LookFor (OPObjectIn value) = 0;
	virtual Bool Find (OPObjectIn value, CountInOut index) = 0;

	OP_DEFINE_CLASS (OPContainer, OPGenericContainer);
};

/****************************************************************/

/*
 *	The OPaC Hash Table is a simple container class which has
 *	relatively fast access times. The storage order cannot be
 *	defined by the user (well, it's a hash table, isn't it ?)
 */

class OP_PUBLIC OPHashTable : public OPContainer
{
	
	struct Element
	{
		Element*		next;					//	next element with same hash value
		OPObjectRef		data;					//	associated data
		
	public:
		Element ()		{ }
		~Element ()		{ }
	};
	
private:
	Count				max_hash;				//	maximum hash value
	Element**			array;					//	array of elements
	
	Count				cache_num;				//	cache: current number
	Count				cache_index;			//	cache: current index in table
	Element**			cache_e_ptr;			//	cache: pointer to prev->next
	Element*			cache_elem;				//	cache: last element
	
protected:
	OPHashTable ();
	virtual ~ OPHashTable ();
	
	virtual void StoreInstance (OPStorageIn storage);	/* OVERRIDDEN */
	virtual void RestoreInstance (OPStorageIn storage);	/* OVERRIDDEN */
	
	virtual void AllocTable (Count);
	virtual void FreeTable ();
	virtual Bool UpdateCache (Count);
	virtual Bool JustInsert (OPObjectIn object);
	
public:
	virtual void Delete ();								/* OVERRIDDEN */
	
	virtual void Initialise (Count max_hash_value);
	
	virtual Bool Insert (OPObjectIn data);				/* OVERRIDDEN */
	virtual Bool Remove (OPObjectIn, Count nth = 0);	/* OVERRIDDEN */
	virtual Bool Remove (Count index);					/* OVERRIDDEN */
	virtual void RemoveAll ();
	
	virtual Bool Get (Count index, OPObjectOut object);	/* OVERRIDDEN */
	virtual Bool GetSimilar (OPObjectIn model, OPObjectOut object,
							 Count nth = 0);			/* OVERRIDDEN */
	
	virtual Bool LookFor (OPObjectIn value);			/* OVERRIDDEN */
	virtual Bool Find (OPObjectIn value, CountInOut);	/* OVERRIDDEN */
	
	OP_DEFINE_CLASS (OPHashTable, OPContainer);
};

/****************************************************************/

/*
 *	The OPaC List class provides an easy way to implement a
 *	sorted list. It can also be used in a less restrictive
 *	way.
 */

class OP_PUBLIC OPList : public OPContainer
{
	friend class OPListIterator;
	
	struct Element
	{
		Element*		next;					//	next element
		Element*		prev;					//	previous element
		
		OPObjectRef		data;					//	associated data
		
	public:
		Element ()		{ }
		~Element ()		{ }
	};
	
private:
	Bool				is_sorted;				//	TRUE => list currently sorted
	Bool				sort_on_insert;			//	TRUE => sort while inserting
	
	Element*			head;					//	pointer to first element
	Element*			tail;					//	pointer to last element
	
	Element*			cache_elem;				//	cached element (if any)
	Count				cache_index;			//	index of element in the list
	
protected:
	OPList ();
	virtual ~ OPList ();
	
	virtual void StoreInstance (OPStorageIn storage);	/* OVERRIDDEN */
	virtual void RestoreInstance (OPStorageIn storage);	/* OVERRIDDEN */
	
	virtual Bool Find (OPObjectIn value, CountInOut, Element* elem);
	
public:
	virtual void Delete ();								/* OVERRIDDEN */
	virtual OPIteratorRef NewIterator ();				/* OVERRIDDEN */
	
	virtual Bool Insert (OPObjectIn data);				/* OVERRIDDEN */
	virtual Bool Remove (OPObjectIn, Count nth = 0);	/* OVERRIDDEN */
	virtual Bool Remove (Count index);					/* OVERRIDDEN */
	virtual void RemoveAll ();							/* OVERRIDDEN */
	
	virtual Bool Get (Count index, OPObjectOut object);	/* OVERRIDDEN */
	virtual Bool GetSimilar (OPObjectIn model, OPObjectOut object,
							 Count nth = 0);			/* OVERRIDDEN */
	
	virtual Bool LookFor (OPObjectIn value);			/* OVERRIDDEN */
	virtual Bool Find (OPObjectIn value, CountInOut);	/* OVERRIDDEN */
	
	virtual Bool IsSorted () const;
	virtual Bool SetInsertMode (Bool sort);
	virtual Bool GetInsertMode () const;
	
	virtual Bool MoveToHead (OPObjectIn value);
	virtual Bool MoveToTail (OPObjectIn value);
	virtual Bool MoveAtPosition (OPObjectIn value, Count index);
	
	virtual Bool PushHead (OPObjectIn value);
	virtual Bool PushTail (OPObjectIn value);
	virtual Bool PopHead (OPObjectOut value);
	virtual Bool PopTail (OPObjectOut value);
	
	OP_DEFINE_CLASS (OPList, OPContainer);
};

/****************************************************************/

/*
 *	Generic data set. A set is a special kind of container
 *	where a key is always associated to the data.
 */

class OP_PUBLIC OPSet : public OPGenericContainer
{
	
protected:
	Count				num_element;			//	# elements in set
	
	Card32				extr_mode;				//	extraction mode
	OPObjectRef			extr_key;				//	key to be used for extraction
	OPObjectRef			extr_data;				//	data to be used for extraction
	
protected:
	OPSet ();
	virtual ~ OPSet ();
	
	virtual void StoreInstance (OPStorageIn storage);	/* OVERRIDDEN */
	virtual void RestoreInstance (OPStorageIn storage);	/* OVERRIDDEN */
	
public:
	virtual void Delete ();								/* OVERRIDDEN */
	
	virtual void SpecifyExtractKey (OPObjectIn key);
	virtual void SpecifyExtractValue (OPObjectIn data);
	virtual Bool SpecifyExtraction (OPArrayIn x);		/* OVERRIDDEN */
	virtual Count TotalCount () const;					/* OVERRIDDEN */
	virtual Bool Get (Count index, OPObjectOut object);	/* OVERRIDDEN */
	
	virtual Bool Insert (OPObjectIn key, OPObjectIn data) = 0;
	virtual Bool Remove (OPObjectIn key, Count nth = 0) = 0;
	virtual Bool Remove (Count index) = 0;
	virtual void RemoveAll () = 0;
	virtual Bool RemoveValue (OPObjectIn value, Count nth = 0) = 0;
	
	virtual Bool Get (Count, OPObjectOut key, OPObjectOut data) = 0;
	virtual Bool GetFromKey (OPObjectIn key, OPObjectOut data, Count nth = 0) = 0;
	virtual Bool GetFromValue (OPObjectIn data, OPObjectOut key, Count nth = 0) = 0;

	OP_DEFINE_CLASS (OPSet, OPGenericContainer);
};

/****************************************************************/

/*
 *	The Dictionary class is used to store key/value pairs. They
 *	are sorted, using the key and can be accessed sequentially.
 */

class OP_PUBLIC OPDict : public OPSet
{
	friend class OPDictIterator;
	
private:
	struct Element
	{
		Element*		next;					//	next element
		Element*		prev;					//	previous element
		
		OPObjectRef		key;					//	key used for sorting
		OPObjectRef		value;					//	value associated to key
		
	public:
		Element ()		{ }
		~Element ()		{ }
	};
	
private:
	Element*			head;					//	first element
	Element*			tail;					//	last element
	
protected:
	OPDict ();
	virtual ~ OPDict ();
	
	virtual void StoreInstance (OPStorageIn storage);	/* OVERRIDDEN */
	virtual void RestoreInstance (OPStorageIn storage);	/* OVERRIDDEN */
	
public:
	virtual void Delete ();								/* OVERRIDDEN */
	virtual OPIteratorRef NewIterator ();				/* OVERRIDDEN */
	
	virtual Bool Insert (OPObjectIn k, OPObjectIn d);	/* OVERRIDDEN */
	virtual Bool Remove (OPObjectIn k, Count = 0);		/* OVERRIDDEN */
	virtual Bool Remove (Count index);					/* OVERRIDDEN */
	virtual void RemoveAll ();							/* OVERRIDDEN */
	virtual Bool RemoveValue (OPObjectIn v, Count = 0);	/* OVERRIDDEN */
	
	virtual Bool Get (Count, OPObjectOut, OPObjectOut);	/* OVERRIDDEN */
	virtual Bool GetFromKey (OPObjectIn key_model, OPObjectOut data,
							 Count nth = 0);			/* OVERRIDDEN */
	virtual Bool GetFromValue (OPObjectIn value_model, OPObjectOut key,
							   Count nth = 0);			/* OVERRIDDEN */

	OP_DEFINE_CLASS (OPDict, OPSet);
};

#endif /* _OPaC_CONTAINER_H_ */

