/*
 *	event.h
 *
 *	This file exports the user event definitions. This is quite portable, but
 *	needs heavy translation by the OPaC event manager.
 *
 *	(C) Copyright 1994-1998, Pierre ARNAUD, OPaC bright ideas
 *		CH-1437 SUSCEVAZ, Switzerland. All rights reserved.
 */

/******************************************************************************

  This file is part of the OPaC class library ("OPaC").

  OPaC is distributed with NO WARRANTY OF ANY KIND.  No author or distributor
  accepts any responsibility for the consequences of using it, or for whether
  it serves any particular purpose or works at all, unless he or she says so
  in writing. Refer to the OPaC Free Public License (the "License") for full
  details.

  Every copy of OPaC must include a copy of the License, normally in a plain
  ASCII text file named PUBLIC. The License grants you the right to copy,
  modify and redistribute OPaC, but only under certain conditions described
  in the License. The copyright notice and this notice must be preserved on
  all copies.

******************************************************************************/

#ifndef	_OPaC_EVENT_H_
#define	_OPaC_EVENT_H_

#ifndef _OPaC_OPEXPORT_H_
#include <OPaC/opexport.h>
#endif

OP_INTERFACE (event.h);

#ifndef	_OPaC_KEY_H_
#include <OPaC/key.h>
#endif

#ifndef	_OPaC_ATOM_H_
#include <OPaC/atom.h>
#endif

/****************************************************************/

/*
 *	Enumeration of the available OPaC event classes. These
 *	take into account only user oriented events. I know
 *	this isn't complete, but it is enough to start working
 *	with.
 */

#define	OP_EVENT_NONE		((Card32)(0))	//	no event - used internally
	
#define	OP_EVENT_KEY_DOWN	((Card32)(1))	//	keyboard: key pressed
#define	OP_EVENT_KEY_REP	((Card32)(2))	//	keyboard: automatic key repetition
#define	OP_EVENT_KEY_UP		((Card32)(3))	//	keyboard: key released
#define	OP_EVENT_KEY_CHANGE	((Card32)(4))	//	keyboard: one shift/modifier changed
	
#define	OP_EVENT_PRESS		((Card32)(5))	//	pointing device: button pressed
#define	OP_EVENT_RELEASE	((Card32)(6))	//	pointing device: button released
#define	OP_EVENT_MOTION		((Card32)(7))	//	pointing device: moved
	
#define	OP_EVENT_IPC_PROTO	((Card32)(8))	//	inter-process communication: protocol request
#define	OP_EVENT_IPC_DATA	((Card32)(9))	//	inter-process communication: data ready
	
#define	OP_EVENT_USER		((Card32)(10))	//	got a user event

/****************************************************************/

/*
 *	Description of the pointing device's set of buttons.
 *	I hope there are enough buttons to accomodate every
 *	acceptable pointing device.
 */

#define	OP_EV_BUTTON_LEFT	(OP_BIT (0))	//	set => left (or only) button pressed
#define	OP_EV_BUTTON_RIGHT	(OP_BIT (1))	//	set => right (or second) button pressed
#define	OP_EV_BUTTON_MIDDLE	(OP_BIT (2))	//	set => middle (front) button pressed
#define	OP_EV_BUTTON_EXTRA	(OP_BIT (3))	//	set => extra button pressed

/****************************************************************/

/*
 *	The OPaC Event record is being represented by a class,
 *	because it has to do some initialization and needs the
 *	safety offered by encapsulation.
 *
 *	There is only one record for any kind of event, thus
 *	simplifying a lot the event handling.
 */

class OP_PUBLIC OPEvent : public OPObject
{
	
private:
	static OPEventRef	current;			//	reference to the current event
	
private:
	Card32				type;				//	event type
	Card32				id;					//	event unique ID
	
	OPWindowRef			window;				//	window in which event occurred
	
	OPRelCoord			x;					//	pointer: horizontal position (wdo relative)
	OPRelCoord			y;					//	pointer: vertical position (wdo relative)
	Card32				buttons;			//	pointer: set of current buttons
	
	Card32				change;				//	specific: change (button, key, etc)
	OPObjectRef			data;				//	specific: data (user data, etc)
	OPAtomRef			name;				//	specific: name (key name, etc)
	Card32				code;				//	specific: code (OEM code, etc)
	
	OPKey				key;				//	keyboard: shift, control & meta keys
	
	Bool				is_dead;			//	TRUE => no more events can be fetched
	
protected:
	OPEvent ();
	virtual ~ OPEvent ();
	
	virtual void StoreInstance (OPStorageIn storage);	/* OVERRIDDEN */
	virtual void RestoreInstance (OPStorageIn storage);	/* OVERRIDDEN */
	
public:
	//	The method used to wait for some event should only be used by the OPaC
	//	Application class !
	
	virtual Bool WaitForEvent ();
	virtual void DispatchEvent ();
	
	//	Accessor members :
	
	Card32 ReturnType () const				{ return this->type; }
	Card32 ReturnID () const				{ return this->id; }
	const OPWindowRef& ReturnWindow () const { return this->window; }
	OPRelCoord ReturnX () const				{ return this->x; }
	OPRelCoord ReturnY () const				{ return this->y; }
	void GetXY (OPRelCoordOut x, OPRelCoordOut y) const	{ x = this->x; y = this->y; }
	Card32 ReturnButtons () const			{ return this->buttons; }
	Card32 ReturnChange () const			{ return this->change; }
	const OPObjectRef& ReturnData () const	{ return this->data; }
	const OPAtomRef& ReturnName () const	{ return this->name; }
	Card32 ReturnCode () const				{ return this->code; }
	const OPKey& ReturnKey () const			{ return this->key; }
	
	//	Global members used for the initialization and for the clean up of the
	//	event manager :
	
	static void Initialise ();
	static void Kill ();
	
	static const OPEventRef& CurrentEvent () { return OPEvent::current; }
	static void HandOverToWindow (OPWindowIn window);
	static void GrabWindowEvents (OPWindowIn window, Bool mode);
	static void PostQuitEvent ();
	
	OP_DEFINE_CLASS (OPEvent, OPObject);
};

#endif /* _OPaC_EVENT_H_ */
