/*
 *	except.h
 *
 *	This file defines the exception handling macros and different
 *	classes used by OPaC to maintain exception associated data.
 *	This will vanish soon !
 *
 *	OPaC does not yet use the ANSI-C++ exceptions, since lots of
 *	compilers do not support these. I hope GNU GCC soon implements
 *	them for the 680x0.
 *	
 *	(C) Copyright 1994-1998, Pierre ARNAUD, OPaC bright ideas
 *		CH-1437 SUSCEVAZ, Switzerland. All rights reserved.
 */

/******************************************************************************

  This file is part of the OPaC class library ("OPaC").

  OPaC is distributed with NO WARRANTY OF ANY KIND.  No author or distributor
  accepts any responsibility for the consequences of using it, or for whether
  it serves any particular purpose or works at all, unless he or she says so
  in writing. Refer to the OPaC Free Public License (the "License") for full
  details.

  Every copy of OPaC must include a copy of the License, normally in a plain
  ASCII text file named PUBLIC. The License grants you the right to copy,
  modify and redistribute OPaC, but only under certain conditions described
  in the License. The copyright notice and this notice must be preserved on
  all copies.

******************************************************************************/

#ifndef	_OPaC_EXCEPT_H_
#define	_OPaC_EXCEPT_H_

#ifndef _OPaC_OPEXPORT_H_
#include <OPaC/opexport.h>
#endif

#ifndef _OPaC_ATOM_H_
#include <OPaC/atom.h>
#endif

OP_INTERFACE (except.h);

class OPException;
class OPExcMemory;
class OPExcObject;
class OPExcTarget;
class OPExcSource;
class OPExcString;

class OPaC_ExcLink;

/****************************************************************/

/*
 *	This macro allows to check the conformity of an object's
 *	class at run-time. It will throw an exception on mismatch.
 */

#define	OP_CLASS_CHECK(x,c)										\
	do {														\
		const OPObjectRef& __o = x;								\
		if (__o.IsInvalid ())							\
			THROW (OPExcObject (OPExcObject::NOT_INITIALISED));	\
		if (! __o->IsKindOf (OP_RUNTIME (c)))					\
			THROW (OPExcObject (OPExcObject::MISMATCH));		\
	} while (0)

#define OP_STRICT_CLASS_CHECK(x,c)								\
	do {														\
		const OPObjectRef& __o = x;								\
		if (__o.IsInvalid ())							\
			THROW (OPExcObject (OPExcObject::NOT_INITIALISED));	\
		if (! __o->IsA (OP_RUNTIME (c)))						\
			THROW (OPExcObject (OPExcObject::MISMATCH));		\
	} while (0)

/****************************************************************/

/*
 *	For every "try", we maintain informations used to jump
 *	back to the corresponding "catch" block, if there is a
 *	"throw" somewhere.
 *	
 *	This class should never be used directly. It is only
 *	provided for the TRY/CATCH/THROW macros.
 */

class OPaC_ExcLink
{
	static OPaC_ExcLink* top_link;				//	pointer to topmost TRY frame
	
protected:
	OPException*		current_exc_inst;		//	current exception instance
	const LatinChar*	current_exc_name;		//	current exception name
	
	OPaC_ExcLink*		prev;					//	previous frame
	
public:
	JumpBuffer			caller;					//	used by setjmp/longjmp
	
public:
	OPaC_ExcLink ();
	~ OPaC_ExcLink ();
	
	Bool Matching (const LatinChar* name);
	OPException* Exception ();
	
	void EndCatch ();
	static void ThrowException (OPException* exc, const LatinChar* name,
								Count line, const LatinChar* file);
};

/****************************************************************/

/*
 *	Macros for handling exceptions. We cannot use the try,
 *	catch and throw keywords, because there are some brain
 *	damaged C++ compilers which don't support these (yet).
 *	
 *	The macros have indeed the same kind of behaviour.
 *
 *	Usage :
 *	
 *	  void
 *	  bar ()
 *	  {
 *	      if (...) THROW (OPExceptionXYZ (...));
 *	      if (...) THROW (OPExceptionABC (...));
 *	  }
 *	  
 *	  void
 *	  foo ()
 *	  {
 *	      TRY {
 *	          bar ();
 *	      }
 *	      
 *	      CATCH (OPExceptionXYZ, exc) {
 *	          ...
 *	      }
 *	      
 *	      AND_CATCH (OPExceptionABC, exc) {
 *	          ...
 *	      }
 *	      
 *	      END_CATCH;
 *	  }
 */                                                          

#define TRY														\
	do {														\
		OPaC_ExcLink _tmp_exc_link;								\
		if (setjmp (_tmp_exc_link.caller) == 0)

#define CATCH(T,e)												\
		else {													\
			if (_tmp_exc_link.Matching (# T)) {					\
				T* e = (T*) _tmp_exc_link.Exception ();			\
				/* catch block comes here */

#define AND_CATCH(T,e)											\
			} else if (_tmp_exc_link.Matching (# T)) {			\
				T* e = (T*) _tmp_exc_link.Exception ();			\
				/* catch block comes here */

#define END_CATCH												\
			} else _tmp_exc_link.EndCatch ();					\
		}														\
	} while (0)

#define THROW(T)												\
	OPaC_ExcLink::ThrowException ( (OPException*) new T, # T,	\
								   __LINE__, __FILE__ )

/****************************************************************/

/*
 *	Base class for exception processing. Every exception
 *	class inherits its properties.
 */

class OP_PUBLIC OPException
{
	friend class OPaC_ExcLink;
	
protected:
	Count				throw_line;				//	line # where throw occurred
	const LatinChar*	throw_file;				//	file name where throw occurred
	
public:
	void* operator new (size_t);				//	for THROW macro only !
	
	Count GetThrowSrcLine () const;
	const LatinChar* GetThrowSrcFile () const;
};

/****************************************************************/

/*
 *	Memory exception class. A more detailed information
 *	about the memory exception is stored in each exception
 *	instance.
 */

class OP_PUBLIC OPExcMemory : public OPException
{
	
public:
	enum Error {
		FULL,									//	memory full (exhausted)
		BUFFER_TOO_SMALL,						//	specified buffer is to small
		NULL_PTR,								//	access to null pointer
		BAD_PTR,								//	access to bad pointer
		FREE									//	access to freed memory
	};
	
protected:
	Error				error;					//	error code
	void*				ptr;					//	faulty pointer (if BAD_PTR)
	
	Count				caller_line;			//	line # where call occurred
	const LatinChar*	caller_file;			//	file name where call occurred
	
public:
	OPExcMemory (Error e, void* p, Count l, const LatinChar* f)
		: error (e), ptr (p), caller_line (l), caller_file (f) { }
	
	Error GetError () const { return this->error; }
	void* GetFaultyPtr () const { return this->ptr; }
	Count GetFaultySrcLine () const { return this->caller_line; }
	const LatinChar* GetFaultySrcFile () const { return this->caller_file; }
};

/****************************************************************/

/*
 *	Object exception class. More information is stored in
 *	the exception instance.
 */

class OP_PUBLIC OPExcObject : public OPException
{
	
public:
	enum Error {
		MISMATCH,								//	object does not match requirements
		NOT_INITIALISED,						//	object has not been initialised
		NOT_FOUND								//	class has not been found
	};
	
protected:
	Error				error;					//	error code
	const OPObject*		object;					//	pointer to faulty object
	
public:
	OPExcObject (Error e, const OPObject* o = 0)
		: error (e), object (o) { }
	
	Error GetError () const { return this->error; }
	const OPObject* GetFaultyObject () const { return this->object; }
};

/****************************************************************/

/*
 *	Target exception class. An access to an invalid target
 *	has been attempted.
 */

class OP_PUBLIC OPExcTarget : public OPException
{
	
protected:
	OPAtomRef			atom;					//	target description
	
public:
	OPExcTarget () : atom ((OPAtom*)(0)) { }
	OPExcTarget (OPAtomIn a) : atom (a) { }
	const OPAtomRef& GetFaultyTarget () const { return this->atom; }
};

/****************************************************************/

/*
 *	Outlet exception class. An access to an invalid outlet
 *	has been attempted.
 */

class OP_PUBLIC OPExcOutlet : public OPException
{
	
protected:
	OPAtomRef			atom;					//	outlet description
	
public:
	OPExcOutlet () : atom ((OPAtom*)(0)) { }
	OPExcOutlet (OPAtomIn a) : atom (a) { }
	const OPAtomRef& GetFaultyOutlet () const { return this->atom; }
};

/****************************************************************/

/*
 *	General access error exception class. An access to an
 *	invalid element in an array has been attempted.
 */

class OP_PUBLIC OPExcAccess : public OPException
{
	
public:
	enum Error {
		INVALID_ACCESS,							//	access declared to be invalid
		OUT_OF_RANGE							//	access is out of range
	};
	
protected:
	Error				error;					//	error description
	
public:
	OPExcAccess (Error e) : error (e) { }
	Error GetError () const { return this->error; }
};

/****************************************************************/

/*
 *	The string exceptions are raised when some invalid
 *	string access takes place (eg. "out of range").
 */

class OP_PUBLIC OPExcString : public OPException
{
	
public:
	enum Error {
		OUT_OF_RANGE							//	access is out of range
	};
	
protected:
	Error				error;					//	error description
public:
	OPExcString (Error e) : error (e) { }
	Error GetError () const { return this->error; }
};

#endif
