/*
 *	rect.h
 *
 *	The rectangle class implements a simple, non-virtual, rectangle. It can be
 *	viewed as a structure.
 *
 *	(C) Copyright 1994-1998, Pierre ARNAUD, OPaC bright ideas
 *		CH-1437 SUSCEVAZ, Switzerland. All rights reserved.
 */

/******************************************************************************

  This file is part of the OPaC class library ("OPaC").

  OPaC is distributed with NO WARRANTY OF ANY KIND.  No author or distributor
  accepts any responsibility for the consequences of using it, or for whether
  it serves any particular purpose or works at all, unless he or she says so
  in writing. Refer to the OPaC Free Public License (the "License") for full
  details.

  Every copy of OPaC must include a copy of the License, normally in a plain
  ASCII text file named PUBLIC. The License grants you the right to copy,
  modify and redistribute OPaC, but only under certain conditions described
  in the License. The copyright notice and this notice must be preserved on
  all copies.

******************************************************************************/

#ifndef	_OPaC_RECT_H_
#define	_OPaC_RECT_H_

#ifndef _OPaC_OPEXPORT_H_
#include <OPaC/opexport.h>
#endif

OP_INTERFACE (rect.h);

/****************************************************************/

class OPRect;

OP_TYPECLASS (OPRect);

/****************************************************************/

/*
 *	The rectangle class provides some easy way to handle
 *	boxes for the graphical operations.
 *	
 *	NB:	OPRect x (10, 10, 19, 19);
 *		x.IsInside (10, 10) => TRUE
 *		x.IsInside (19, 19) => TRUE
 *		x.GetDx ()          => 10 points
 *		
 *		OPRect y (10, 10, 10, 10);
 *		y.IsInside (10, 10) => TRUE
 *		y.GetDx ()          => 1 point
 *	
 *	This policy has been used so that it will be easy to
 *	use with the Grail classes. A box may have the size of
 *	one pixel... but it won't be considered to be empty.
 */

class OP_PUBLIC OPRect
{
	
private:
	OPCoord				x1, y1, x2, y2;			//	bouding box of rectangle
	Bool				is_empty;				//	TRUE => empty box
	
public:
	OPRect ()
		{ this->x1 = this->y1 = this->x2 = this->y2 = (OPCoord)(0);
		  this->is_empty = TRUE; }
	
	OPRect (OPRectIn box, OPRelCoord margin = 0)
		{ OPRelCoord dx = (OPRelCoord)(box.DX ());
		  OPRelCoord dy = (OPRelCoord)(box.DY ());
		  if ( (box.is_empty) || (dx < margin) || (dy < margin) ) {
		      this->x1 = this->y1 = this->x2 = this->y2 = (OPCoord)(0);
		      this->is_empty = TRUE;
		  } else {
		      this->x1 = box.x1 + margin; this->y1 = box.y1 + margin;
		      this->x2 = box.x2 - margin; this->y2 = box.y2 - margin;
		      this->is_empty = FALSE; } }
	
	OPRect (OPCoord x1, OPCoord y1, OPCoord x2, OPCoord y2)
		{ if ( (x2 < x1) || (y2 < y1) ) {
		      this->x1 = this->y1 = this->x2 = this->y2 = (OPCoord)(0);
		      this->is_empty = TRUE;
		  } else {
		      this->x1 = x1; this->y1 = y1; this->x2 = x2; this->y2 = y2;
		      this->is_empty = FALSE; } }
	
	OPRect& operator  = (OPRectIn box)
		{ this->x1 = box.x1; this->y1 = box.y1;
		  this->x2 = box.x2; this->y2 = box.y2;
		  this->is_empty = box.is_empty; return *this; }
	
	OPRect& operator += (OPRectIn box);
	OPRect& operator *= (OPRectIn box);
	
	OPRect& InsertPoint (OPCoord x, OPCoord y)
		{ if (this->is_empty) {
			  this->x1 = this->x2 = x;
			  this->y1 = this->y2 = y;
			  this->is_empty = FALSE;
		  } else {
			  if (x < this->x1) this->x1 = x;
			  if (y < this->y1) this->y1 = y;
			  if (x > this->x2) this->x2 = x;
			  if (y > this->y2) this->y2 = y;
		  }
		  return *this; }
	
	OPRect& MakeEmpty ()
		{ this->x1 = this->y1 = this->x2 = this->y2 = (OPCoord)(0);
		  this->is_empty = TRUE; return *this; }
	
	OPRect& SetOrigin (OPCoord x, OPCoord y)
		{ this->x2 += x - this->x1; this->y2 += y - this->y1;
		  this->x1 = x; this->y1 = y; return *this; }
	
	OPRect& MoveOrigin (OPCoord dx, OPCoord dy)
		{ this->x1 += dx; this->x2 += dx;
		  this->y1 += dy; this->y2 += dy; return *this; }
	
	OPRect& SetSize (OPCoord dx, OPCoord dy)
		{ this->x2 = this->x1 + dx - 1;
		  this->y2 = this->y1 + dy - 1;
		  this->is_empty = ((dx == (OPCoord)(0)) || (dy == (OPCoord)(0))) ? TRUE : FALSE;
		  return *this; }
	
	OPRect& ChangeSize (OPCoord dx, OPCoord dy)
		{ dx = (this->is_empty) ? dx : (this->x2 - this->x1 + 1 + dx);
		  dy = (this->is_empty) ? dy : (this->y2 - this->y1 + 1 + dy);
		  if (dx < 0) dx = 0;
		  if (dy < 0) dy = 0;
		  return this->SetSize (dx, dy); }
	
	Bool operator == (OPRectIn box) const
		{ return ( (this->is_empty && box.is_empty)
				|| ( (this->x1 == box.x1) && (this->y1 == box.y1)
				  && (this->x2 == box.x2) && (this->y2 == box.y2) ) )
				 ? TRUE : FALSE; }
		
	Bool operator != (OPRectIn box) const
		{ return ( (this->is_empty != box.is_empty)
				|| (this->x1 != box.x1) || (this->y1 != box.y1)
				|| (this->x2 != box.x2) || (this->y2 != box.y2) )
				 ? TRUE : FALSE; }
	
	Bool IsInside (OPCoord x, OPCoord y) const
		{ return ( (this->is_empty) || (x < this->x1) || (y < this->y1)
				|| (x > this->x2) || (y > this->y2) )
				 ? FALSE : TRUE; }
	
	Bool DoOverlap (OPRectIn rect) const
		{ return ( (this->is_empty == FALSE) && (rect.is_empty == FALSE)
		        && (this->x1 <= rect.x2) && (this->x2 >= rect.x1)
		        && (this->y1 <= rect.y2) && (this->y2 >= rect.y1) )
		         ? TRUE : FALSE; }
	
	Bool IsEmpty () const	{ return this->is_empty; }
	
	OPCoord X1 () const		{ return this->x1; }
	OPCoord Y1 () const		{ return this->y1; }
	OPCoord X2 () const		{ return this->x2; }
	OPCoord Y2 () const		{ return this->y2; }
	
	OPCoord& X1 ()			{ return this->x1; }
	OPCoord& Y1 ()			{ return this->y1; }
	OPCoord& X2 ()			{ return this->x2; }
	OPCoord& Y2 ()			{ return this->y2; }
	OPCoord DX () const		{ return this->is_empty ? (OPCoord)(0) : (this->x2 - this->x1 + 1); }
	OPCoord DY () const		{ return this->is_empty ? (OPCoord)(0) : (this->y2 - this->y1 + 1); }
	
	friend OPStorageIn operator << (OPStorageIn, OPRectIn);
	friend OPStorageIn operator >> (OPStorageIn, OPRectOut);
};

/****************************************************************/

#endif
