/*
 *	string.h
 *
 *	This file defines the string class.
 *
 *	(C) Copyright 1994-1997, Pierre ARNAUD, OPaC bright ideas
 *		CH-1437 SUSCEVAZ, Switzerland. All rights reserved.
 */

/******************************************************************************

  This file is part of the OPaC class library ("OPaC").

  OPaC is distributed with NO WARRANTY OF ANY KIND.  No author or distributor
  accepts any responsibility for the consequences of using it, or for whether
  it serves any particular purpose or works at all, unless he or she says so
  in writing. Refer to the OPaC Free Public License (the "License") for full
  details.

  Every copy of OPaC must include a copy of the License, normally in a plain
  ASCII text file named PUBLIC. The License grants you the right to copy,
  modify and redistribute OPaC, but only under certain conditions described
  in the License. The copyright notice and this notice must be preserved on
  all copies.

******************************************************************************/

#ifndef	_OPaC_STRING_H_
#define	_OPaC_STRING_H_

#ifndef _OPaC_OPEXPORT_H_
#include <OPaC/opexport.h>
#endif

OP_INTERFACE (string.h);

/****************************************************************/

extern inline Bool operator == (const LatinChar* a, OPStringIn b) { return (b == a) ? TRUE : FALSE; }
extern inline Bool operator == (const UnicodeChar* a, OPStringIn b) { return (b == a) ? TRUE : FALSE; }
extern inline Bool operator != (OPStringIn a, const LatinChar* b) { return (a == b) ? FALSE : TRUE; }
extern inline Bool operator != (const LatinChar* a, OPStringIn b) { return (a == b) ? FALSE : TRUE; }
extern inline Bool operator != (OPStringIn a, const UnicodeChar* b) { return (a == b) ? FALSE : TRUE; }
extern inline Bool operator != (const UnicodeChar* a, OPStringIn b) { return (a == b) ? FALSE : TRUE; }

/****************************************************************/


/*
 *	The OPString class replaces the "char*" and "char[]"
 *	types in lots of places. It has the advantage to be
 *	OPaC conformant : it can be put into a storage and
 *	it can be manipulated through the network.
 */

class OP_PUBLIC OPString : public OPText
{
	friend class OPAtom;
	
	union AnyTextUnion
	{
		LatinChar*			latin;				//	pointer to ISO Latin-1 text
		UnicodeChar*		unicode;			//	pointer to 16-bit Unicode text
		const LatinChar*	solid_latin;		//	like latin, but points to external text
		const UnicodeChar*	solid_unicode;		//	like unicode, but points to external text
	};
	
	enum AnyTextType
	{
		TEXT_IS_LATIN,							//	union field "latin"
		TEXT_IS_UNICODE,						//	union field "unicode"
		TEXT_IS_SOLID_LATIN,					//	union field "solid_latin"
		TEXT_IS_SOLID_UNICODE					//	union field "solid_unicode"
	};
	
protected:
	AnyTextUnion	text;						//	text buffer or external text
	AnyTextType		type;						//	type of "text" field
	
	Count			len;						//	text length (0..n)
	Count			buffer_len;					//	buffer length (if any)
	
protected:
	OPString ();
	virtual ~OPString ();
	
	virtual void StoreInstance (OPStorageIn storage);	/* OVERRIDDEN */
	virtual void RestoreInstance (OPStorageIn storage);	/* OVERRIDDEN */
	
	virtual const LatinChar* ReturnQuickLatin () const;	/* OVERRIDDEN */
	virtual const UnicodeChar* ReturnQuickUnicode () const;/* OVERRIDDEN */
	
	virtual void FreeMemory ();
	virtual void AllocMemory (AnyTextType type, Count len);
	virtual void ResizeMemory (Count len);
	
public:
	virtual OPStringRef ReturnString () const;			/* OVERRIDDEN */
	virtual Bool SetString (OPStringIn text);			/* OVERRIDDEN */
	virtual Bool GetString (OPStringOut text) const;
	
	virtual Bool SetString (const LatinChar* text, Count len = 0);
	virtual Bool SetString (const UnicodeChar* text, Count len = 0);
	virtual Bool SetString (OPStringIn text, Count len);
	
	virtual Bool SetSolidString (const LatinChar* text);
	virtual Bool SetSolidString (const UnicodeChar* text);
	
	virtual Bool InsertString (Count pos, const LatinChar* text, Count len = 0);
	virtual Bool InsertString (Count pos, const UnicodeChar* text, Count len = 0);
	virtual Bool InsertString (Count pos, OPStringIn text, Count len = 0);
	
	virtual Bool AppendString (const LatinChar* text, Count len = 0);
	virtual Bool AppendString (const UnicodeChar* text, Count len = 0);
	virtual Bool AppendString (OPStringIn text, Count len = 0);
	
	virtual Bool ClearString ();
	virtual Bool DeleteString (Count pos, Count num);
	
    virtual Count CountCharacters () const;
	virtual Bool IsDefaultLatin () const;
	virtual Bool IsDefaultUnicode () const;
	virtual const LatinChar* AccessLatin () const;
	virtual const UnicodeChar* AccessUnicode () const;
    
    virtual void GetCharacter (Count pos, LatinCharOut character) const;
    virtual void GetCharacter (Count pos, UnicodeCharOut character) const;
    
    virtual void ConvertToUpper ();
    virtual void ConvertToLower ();
    
    virtual void ConvertToLatin ();
    virtual void ConvertToUnicode ();
    
    virtual Bool GetStringCopy (LatinChar* buffer, Count max);
    virtual Bool GetStringCopy (UnicodeChar* buffer, Count max);
    
    virtual Bool ExtractStringCopy (LatinChar* buffer, Count pos, Count num);
    virtual Bool ExtractStringCopy (UnicodeChar* buffer, Count pos, Count num);
	
	virtual Bool Find (UnicodeChar what, CountOut pos, int dir = 1) const;
	virtual Bool Find (OPStringIn what, CountOut pos, int dir = 1) const;
	
	virtual Bool MakeLeft (Count num, OPStringOut text) const;
	virtual Bool MakeMiddle (Count pos, Count num, OPStringOut text) const;
	virtual Bool MakeRight (Count num, OPStringOut text) const;
    
    //	Overridden methods, inherited from the OPObject class :
    
	virtual void Delete ();								/* OVERRIDDEN */
	virtual Count ComputeHashValue (Count num) const;	/* OVERRIDDEN */
	virtual int CompareWith (OPObjectIn object) const;	/* OVERRIDDEN */
	virtual int CompareWith (const LatinChar* s) const;
	virtual int CompareWith (const UnicodeChar* s) const;
	
	static UnicodeChar LatinToUnicode (LatinChar c);
	static LatinChar UnicodeToLatin (UnicodeChar c);
	static LatinChar UnicodeToLatin (const UnicodeChar* c, CountInOut num);
	
	OP_DEFINE_CLASS_SPECIAL_REF (OPString, OPText);
};

/****************************************************************/

extern inline Bool
operator == (OPStringIn a, const LatinChar* b)
{ return (a->CompareWith (b) == OP_COMP_THIS_SAME) ? TRUE : FALSE; }

extern inline Bool
operator == (OPStringIn a, const UnicodeChar* b)
{ return (a->CompareWith (b) == OP_COMP_THIS_SAME) ? TRUE : FALSE; }

/****************************************************************/

inline OPString*
OPStringRef::CheckAndResolve () const
{
	return (OPString*)((this->object) ? this->object->CheckAndResolve ()
									  : this->ReturnRuntime ()->ResolveNull ());
}

inline OPString*
OPStringRef::Resolve () const
{
	return (OPString*)((this->object) ? this->object->Resolve () : 0);
}

inline OPString*
OPStringRef::operator -> () const
{
	return (OPString*)(this->CheckAndResolve ());
}

/****************************************************************/

#endif /* _OPaC_STRING_H_ */
