/*
 *	boxview.cxx
 *
 *	The Box View is used to hold a set of views, which will be surrounded
 *	by a frame and a title (which may both be disabled).
 *
 *	(C) Copyright 1994-1998, Pierre ARNAUD, OPaC bright ideas
 *		CH-1437 SUSCEVAZ, Switzerland. All rights reserved.
 */

/******************************************************************************

  This file is part of the OPaC class library ("OPaC").

  OPaC is distributed with NO WARRANTY OF ANY KIND.  No author or distributor
  accepts any responsibility for the consequences of using it, or for whether
  it serves any particular purpose or works at all, unless he or she says so
  in writing. Refer to the OPaC Free Public License (the "License") for full
  details.

  Every copy of OPaC must include a copy of the License, normally in a plain
  ASCII text file named PUBLIC. The License grants you the right to copy,
  modify and redistribute OPaC, but only under certain conditions described
  in the License. The copyright notice and this notice must be preserved on
  all copies.

******************************************************************************/

#include <OPaC/boxview.h>
#include <OPaC/texttag.h>
#include <OPaC/grail.h>
#include <OPaC/lookman.h>
#include <OPaC/view.h>
#include <OPaC/new.h>
#include <OPaC/pubmacro.h>

/****************************************************************/

OP_IMPLEMENTATION (boxview.h);
OP_IMPLEMENT_CLASS (OPBoxView);

/****************************************************************/

OP_PUBLISH_MODIFIER (OPBoxView, "Title", GetTitle, SetTitle, "Box title", OPObject, OPString);
OP_PUBLISH_MODIFUNC (OPBoxView, "TitleEnable", ReturnTitleEnabled, EnableTitle, "Title enabled", Bool, Bool);
OP_PUBLISH_MODIFUNC (OPBoxView, "FrameEnable", ReturnFrameEnabled, EnableFrame, "Frame enabled", Bool, Bool);

/****************************************************************/

/*
 *	The default box view has no title. An invalid title is
 *	considered to be no title at all.
 */

OPBoxView::OPBoxView () : title (0)
{
	OP_METHOD (OPBoxView, 0);
	
	this->enable_title = TRUE;
	this->enable_frame = TRUE;
	
	this->background_color = OP_COLOR_NONE;
	this->frame_color      = OP_COLOR_NONE;
}



/*
 *	The title will be freed automagically if nobody uses
 *	it.
 */

OPBoxView::~OPBoxView ()
{
	OP_METHOD (~OPBoxView, 0);
}



/*
 *	Prepare the box view for deletion.
 */

void
OPBoxView::Delete ()
{
	OP_METHOD (Delete, 0);
	
	this->title = 0;
	
	super::Delete ();
}



/*
 *	Save the box view : this essentially saves its geometry
 *	and its title.
 */

void
OPBoxView::StoreInstance (OPStorageIn storage)
{
	OP_METHOD (StoreInstance, 1);
	OP_ARG (1, OPStorageIn, storage);
	
	super::StoreInstance (storage);

	OPLookManRef look = OPLookMan::Current ();
	OPAtomRef    name;
	
	storage << this->title;
	storage << this->enable_title << this->enable_frame;
	
	look->GetColorName (this->background_color, name);	storage << name;
	look->GetColorName (this->frame_color, name);		storage << name;
	
	storage << OPStorageSync::mark;
}



/*
 *	Restore the box view : this restores its geometry and
 *	its title.
 */

void
OPBoxView::RestoreInstance (OPStorageIn storage)
{
	OP_METHOD (RestoreInstance, 1);
	OP_ARG (1, OPStorageIn, storage);
	
	super::RestoreInstance (storage);
	
	OPLookManRef look = OPLookMan::Current ();
	OPAtomRef    name;
	
	storage >> this->title;
	storage >> this->enable_title >> this->enable_frame;
	
	storage >> name;	this->background_color = look->FindColor (name);
	storage >> name;	this->frame_color      = look->FindColor (name);
	
	storage >> OPStorageSync::mark;
}



/*
 *	This casting function is needed by the OPaC extension
 *	class OPxTitleWidget.
 */

OPView*
OPBoxView::ReturnObjectAsView ()
{
	return this;
}


/*
 *	Return the box view's minimal size. This will be the
 *	minimal size of the title. This does not take into
 *	account the objects pasted into the view.
 */

void
OPBoxView::GetMinSize (OPCoordOut dx, OPCoordOut dy) const
{
	OP_METHOD (GetMinSize, 2);
	OP_ARG (1, OPCoordOut, dx);
	OP_ARG (2, OPCoordOut, dy);
	
	OPCoord dx1 = 0;
	OPCoord dy1 = 0;
	OPCoord dx2 = 0;
	OPCoord dy2 = 0;
	
	if ( (this->enable_title)
	  && (this->title.IsValid ()) ) {
		this->title->GetMinSize (dx1, dy1);
	}
	
	super::GetMinSize (dx2, dy2);
	
	dx = dx1 > dx2 ? dx1 : dx2;
	dy = dy1 + dy2;
	
	if (this->enable_frame) {
		dx += 4;
		dy += 4;
	}
}



/*
 *	Return the box view's maximum size. Return an arbitrary
 *	huge value.
 */

void
OPBoxView::GetMaxSize (OPCoordOut dx, OPCoordOut dy) const
{
	OP_METHOD (GetMinSize, 2);
	OP_ARG (1, OPCoordOut, dx);
	OP_ARG (2, OPCoordOut, dy);
	
	super::GetMaxSize (dx, dy);
}


/*
 *	Allocates a text tag if this hasn't been done before.
 */

void
OPBoxView::SetupTextTag ()
{
	OP_METHOD (SetupTextTag, 0);
	
	if (this->title.IsInvalid ()) {
		this->title = OP_NEW_FROM_ZONE (OPaC_iface_zone, OPTextTag);
		this->title->SetParent (this);
	}
}



/*
 *	Called whenever the geometry will change.
 */

void
OPBoxView::ChangeGeometry (OPRectIn box)
{
	OP_METHOD (ChangeGeometry, 3);
	OP_ARG (1, OPRectIn, box);
	
	super::ChangeGeometry (box);
	this->UpdateTextTag ();
}



/*
 *	Update the text tag geometry, whenever the box view
 *	gets moved or resized.
 */

void
OPBoxView::UpdateTextTag ()
{
	OP_METHOD (UpdateTextTag, 0);
	
	if (this->title.IsValid ()) {
		
		OPCoord dx = 0;
		OPCoord dy = 0;
		
		this->title->GetMinSize (dx, dy);
		
		dx += 4;
		dy += 2;
		
		OPCoord x1   = this->box.X1 () + (this->box.DX () - dx) / 2;
		OPCoord y1   = this->box.Y1 ();
		OPCoord x2   = x1 + dx - 1;
		OPCoord y2   = y1 + dy - 1;
		OPRect  rect = OPRect (x1, y1, x2, y2);
		
		this->title->SetBox (rect);
	}
}

Bool
OPBoxView::EnableTitle (Bool yes)
{
	OP_METHOD (EnableTitle, 1);
	OP_ARG (1, Bool, yes);
	
	this->enable_title = yes;
	
	return TRUE;
}

Bool
OPBoxView::EnableFrame (Bool yes)
{
	OP_METHOD (EnableFrame, 1);
	OP_ARG (1, Bool, yes);
	
	this->enable_frame = yes;
	
	return TRUE;
}


Bool
OPBoxView::ReturnTitleEnabled () const
{
	OP_METHOD (ReturnTitleEnabled, 0);
	return this->enable_title;
}


Bool
OPBoxView::ReturnFrameEnabled () const
{
	OP_METHOD (ReturnFrameEnabled, 0);
	return this->enable_frame;
}


/*
 *	Define the box view title's font.
 */

Bool
OPBoxView::SetFont (OPGrailFontIn font)
{
	OP_METHOD (SetFont, 1);
	OP_ARG (1, OPGrailFontIn, font);
	
	this->SetupTextTag ();
	this->title->SetFont (font);
	this->UpdateTextTag ();
	
	return TRUE;
}

Bool
OPBoxView::GetFont (OPGrailFontOut font) const
{
	OP_METHOD (GetFont, 1);
	OP_ARG (1, OPGrailFontOut, font);
	
	if (this->title.IsValid ()) {
		this->title->GetFont (font);
		return TRUE;
	}
	
	return FALSE;
}



/*
 *	Define the box view title's contents.
 */

Bool
OPBoxView::SetTitle (OPStringIn text)
{
	OP_METHOD (SetTitle, 1);
	OP_ARG (1, OPStringIn, text);
	
	this->SetupTextTag ();
	this->title->SetText (text);
	this->UpdateTextTag ();
	
	return TRUE;
}

Bool
OPBoxView::GetTitle (OPStringOut text) const
{
	OP_METHOD (GetTitle, 1);
	OP_ARG (1, OPStringOut, text);
	
	if (this->title.IsValid ()) {
		this->title->GetText (text);
		return TRUE;
	}
	
	return FALSE;
}


/*
 *	Define the background color and the frame color. If the
 *	specified color is OP_COLOR_NONE, then the default box
 *	view colors will be used (no background color and a 3D
 *	frame).
 */

Bool
OPBoxView::SetBackgroundColor (OPLookColor color)
{
	OP_METHOD (SetBackgroundColor, 1);
	OP_ARG (1, OPLookColor, color);
	
	this->background_color = color;
	
	return TRUE;
}

Bool
OPBoxView::SetFrameColor (OPLookColor color)
{
	OP_METHOD (SetFrameColor, 1);
	OP_ARG (1, OPLookColor, color);
	
	this->frame_color = color;
	
	return TRUE;
}

OPLookColor
OPBoxView::ReturnBackgroundColor () const
{
	OP_METHOD (ReturnBackgroundColor, 0);
	return this->background_color;
}

OPLookColor
OPBoxView::ReturnFrameColor () const
{
	OP_METHOD (ReturnFrameColor, 0);
	return this->frame_color;
}



/*
 *	Draw the box view. This will automatically draw the
 *	text tag.
 */

void
OPBoxView::Draw (OPGrailPortIn port, OPRectIn rect, OPDrawOp what)
{
	OP_METHOD (Draw, 3);
	OP_ARG (1, OPGrailPortIn, port);
	OP_ARG (2, OPRectIn, rect);
	OP_ARG (3, OPDrawOp, what);
	
	const OPAdornerRef& adorner = port->ReturnAdorner ();
	
	OPCoord delta = this->enable_title ? 6 : 0;
	OPRect  box   = this->box;
	
	box.Y1 () += delta;
	
	if ( (what & OP_DRAW_BACKGROUND)
	  && (this->background_color != OP_COLOR_NONE) ) {
		
		what &= ~ OP_DRAW_BACKGROUND;
		
		adorner->FillRect (box, this->background_color);
	}
	
	super::Draw (port, rect, what);
	
	if (what & OP_DRAW_OBJECT) {
		
		if (this->enable_frame) {
			
			//	Draw the frame (if it is enabled).
			
			if (this->frame_color == OP_COLOR_NONE) {
				adorner->DrawFrame (box);
			} else {
				adorner->DrawRect (box, this->frame_color);
			}
		}
		
		if (this->enable_title) {
			
			//	Draw the title (if it is enabled).
			
			if (this->title.IsValid ()) {
				
				OPCoord x  = 0;
				OPCoord y  = 0;
				OPCoord dx = 0;
				OPCoord dy = 0;
				
				this->title->GetSize (dx, dy);
				this->title->GetOrigin (x, y);
				
				adorner->SetForeColor (OP_COLOR_BACKGROUND);
				port->FillRect (x, y, x + dx - 1, y + dy - 1);
				
				this->title->Draw (port, rect, what);
			}
		}
	}
}

