/*
 *	datalog.cxx
 *
 *	Data logger : this is a simple way to store lots of data
 *	(which may be stored on disk when needed). The data can
 *	be of the same type or variable.
 *
 *	(C) Copyright 1994-1998, Pierre ARNAUD, OPaC bright ideas
 *		CH-1437 SUSCEVAZ, Switzerland. All rights reserved.
 */

/******************************************************************************

  This file is part of the OPaC class library ("OPaC").

  OPaC is distributed with NO WARRANTY OF ANY KIND.  No author or distributor
  accepts any responsibility for the consequences of using it, or for whether
  it serves any particular purpose or works at all, unless he or she says so
  in writing. Refer to the OPaC Free Public License (the "License") for full
  details.

  Every copy of OPaC must include a copy of the License, normally in a plain
  ASCII text file named PUBLIC. The License grants you the right to copy,
  modify and redistribute OPaC, but only under certain conditions described
  in the License. The copyright notice and this notice must be preserved on
  all copies.

******************************************************************************/

#include <OPaC/array.h>
#include <OPaC/convert.h>
#include <OPaC/datalog.h>
#include <OPaC/new.h>
#include <OPaC/targetmacro.h>
#include <OPaC/values.h>

#include <string.h>

/****************************************************************/

OP_IMPLEMENTATION (datalog.h);
OP_IMPLEMENT_CLASS (OPDataLog);

/****************************************************************/

OP_TARGET_ACTOR (OPDataLog, FlushCache);
OP_TARGET_ACTOR (OPDataLog, FlushCacheAndAutoInsertFromSource);
OP_TARGET_ACTOR (OPDataLog, AutoInsertFromSource);

/****************************************************************/

/*
 *	Setup the data logger.
 */

OPDataLog::OPDataLog ()
{
	OP_METHOD (OPDataLog, 0);
	
	this->type = OP_TYPE_NONE;
	this->simple.byte_array = 0;
	this->complex_array = OPArrayNull;
	
	this->index = 0;
	this->array_size = 0;
	this->cyclic_size = 0;
	
	this->is_cyclic = FALSE;
	this->is_simple = FALSE;
}


/*
 *	Clean-up the data logger.
 */

OPDataLog::~OPDataLog ()
{
	OP_METHOD (~OPDataLog, 0);

	if (this->type != OP_TYPE_NONE) {
		this->InitLog (OP_TYPE_NONE);
	}
}



/*
 *	Prepare the logger for deletion.
 */

void
OPDataLog::Delete ()
{
	OP_METHOD (Delete, 0);
	
	OPxSmartSource::Cleanup ();
	
	if (this->type != OP_TYPE_NONE) {
		this->InitLog (OP_TYPE_NONE);
	}
	
	super::Delete ();
}


/*
 *	Save the log's parameters. Surprisingly, this will not save
 *	the log's data. Use persistent logging if you want to keep
 *	a trace of the logged data.
 */

void
OPDataLog::StoreInstance (OPStorageIn storage)
{
	OP_METHOD (StoreInstance, 1);
	OP_ARG (1, OPStorageIn, storage);
	
	super::StoreInstance (storage);
	OPxSmartSource::StoreInstance (storage);
	
	storage << this->type << this->cyclic_size;
	storage << OPStorageSync::mark;
}



/*
 *	Restore the log's parameters. This will not restore the
 *	log's data. See StoreInstance.
 */

void
OPDataLog::RestoreInstance (OPStorageIn storage)
{
	OP_METHOD (RestoreInstance, 1);
	OP_ARG (1, OPStorageIn, storage);
	
	super::RestoreInstance (storage);
	OPxSmartSource::RestoreInstance (storage);
	
	OPSimpleType type = OP_TYPE_NONE;
	Card32       size = 0;
	
	storage >> type >> size;
	storage >> OPStorageSync::mark;
	
	this->InitLog (type, size);
}


/*
 *	Grow the size of the current data log. This will be called
 *	only if the data type corresponds to a simple one.
 */

void
OPDataLog::GrowSimpleArray (Card32 num)
{
	OP_METHOD (GrowSimpleArray, 1);
	OP_ARG (1, Card32, num);
	
	Byte* current = this->simple.byte_array;
	Byte* larger  = OP_ALLOC_ARRAY (Byte, num * this->simple_size);
	
	if (current) {
		memcpy (larger, current, num * this->simple_size);
		OP_FREE (current);
	}
	
	this->array_size = num;
	this->simple.byte_array = larger;
}


/*
 *	Automatically insert data from the smart source.
 */

void
OPDataLog::AutoInsertFromSource ()
{
	OP_METHOD (AutoInsertFromSource, 0);
	OPValueRef value;
	
	OPxSmartSource::GetNthValue (0, value);
	this->InsertValue (value);
}


/*
 *	Flush the cache before inserting (and therefore generating
 *	a new sample from the smart source) data.
 */

void
OPDataLog::FlushCacheAndAutoInsertFromSource ()
{
	OP_METHOD (FlushCacheAndAutoInsertFromSource, 0);
	
	this->FlushCache ();
	this->AutoInsertFromSource ();
}


/*
 *	Compute the size of the currently selected simple type.
 */

void
OPDataLog::ComputeSimpleSize ()
{
	OP_METHOD (ComputeSimpleSize, 0);
	
	switch (type) {
		case OP_TYPE_NONE:
			this->simple_size = 0;
			break;
		case OP_TYPE_BOOL:
			this->simple_size = sizeof (Bool);
			break;
		case OP_TYPE_CARD:
			this->simple_size = sizeof (Card32);
			break;
		case OP_TYPE_INT:
			this->simple_size = sizeof (Int32);
			break;
		case OP_TYPE_REAL:
			this->simple_size = sizeof (Real);
			break;
		default:
			this->simple_size = 0;
			break;
	}
}



/*
 *	Insert some data into the array. The size of the data
 *	and the offset in the array depend on the simple type
 *	associated to it.
 */

Bool
OPDataLog::InsertAnyType (void* any)
{
	OP_METHOD (InsertAnyType, 1);
	OP_ARG (1, void*, any);
	
	Count pos = this->index++;
	
	if (this->is_cyclic) {
		if (pos >= this->cyclic_size) {
			pos %= this->cyclic_size;
		}
	}
	
	if (pos >= this->array_size) {
		Count max = (this->is_cyclic) ? (this->cyclic_size) : (pos+50);
		this->GrowSimpleArray (max);
	}
	
	Byte* source = (Byte*)(any);
	Byte* dest   = this->simple.byte_array + pos * this->simple_size;
	Count size   = this->simple_size;
	
	while (size--) {
		*dest++ = *source++;
	}

	return TRUE;
}



/*
 *	Initialise the data logger. This will free any previously
 *	acquired data.
 */


void
OPDataLog::InitLog (OPSimpleType type, Count cyclic_size)
{
	OP_METHOD (InitLog, 2);
	OP_ARG (1, OPSimpleType, type);
	OP_ARG (2, Count, cyclic_size);
	
	switch (type) {
		case OP_TYPE_NONE:
		case OP_TYPE_BOOL:
		case OP_TYPE_CARD:
		case OP_TYPE_INT:
		case OP_TYPE_REAL:
			this->is_simple = TRUE;
			break;

		default:
			type = OP_TYPE_OBJECT;
			this->is_simple = FALSE;
			break;
	}
	
	this->type = type;
	
	if (this->complex_array.IsValid ()) {
		this->complex_array->Delete ();
		this->complex_array = OPArrayNull;
	}
	
	if (this->simple.byte_array) {
		OP_FREE (this->simple.byte_array);
		this->simple.byte_array = 0;
	}
	
	this->index = 0;
	this->array_size = 0;
	this->cyclic_size = cyclic_size;
	this->is_cyclic = (cyclic_size) ? TRUE : FALSE;
	
	if (this->is_simple) {
		this->ComputeSimpleSize ();
	} else {
		this->complex_array = OP_NEW (OPArray);
	}
}


/*
 *	Return the simple (or complex) type used by the logger.
 */

OPSimpleType
OPDataLog::ReturnType () const
{
	OP_METHOD (ReturnType, 0);
	return this->type;
}


/*
 *	Insert a value into the data logger. This will only be
 *	possible if the value is compatible with the logging
 *	type.
 */

Bool
OPDataLog::InsertValue (OPValueIn value)
{
	OP_METHOD (InsertValue, 1);
	OP_ARG (1, OPValueIn, value);
	
	if ( (this->type == OP_TYPE_OBJECT)
	  && (this->complex_array.IsValid ()) ) {
		
		Count pos = this->index++;
		
		if (this->is_cyclic) {
			if (pos >= this->cyclic_size) {
				pos %= this->cyclic_size;
			}
		}
		
		if (pos >= this->array_size) {
			Count max = (this->is_cyclic) ? (this->cyclic_size) : (pos+50);
			this->complex_array->SetSize (max);
		}
		
		this->complex_array[pos] = value;
		return TRUE;
	}
	
	
	//	Convert the value to the appropriate simple type and try
	//	it that way. Hope this works...
	
	if ( (value.IsValid ())
	  && (value->ReturnType () <= OP_TYPE_REAL) ) {
		
		switch (this->type) {
			case OP_TYPE_BOOL:	return this->InsertBool (value);
			case OP_TYPE_CARD:	return this->InsertCard32 (value);
			case OP_TYPE_INT:	return this->InsertInt32 (value);
			case OP_TYPE_REAL:	return this->InsertReal (value);
		}
	}
	
	return FALSE;
}


/*
 *	Return the nth value stored in the data logger.
 */

Bool
OPDataLog::GetNthValue (Count index, OPValueOut value) const
{
	OP_METHOD (GetNthValue, 2);
	OP_ARG (1, Count, index);
	OP_ARG (2, OPValueOut, value);

	if (index < this->index) {
		
		if (this->is_cyclic) {
			index %= this->cyclic_size;
		}
		
		if (index < this->array_size) {
			
			if (this->is_simple) {
				switch (this->type) {
					case OP_TYPE_BOOL:	value = this->simple.bool_array[index]; break;
					case OP_TYPE_INT:	value = this->simple.int_array[index];  break;
					case OP_TYPE_CARD:	value = this->simple.card_array[index]; break;
					case OP_TYPE_REAL:	value = this->simple.real_array[index]; break;
					default:			value = OPValueNull;					break;
				}
			} else {
				value = this->complex_array[index];
			}
			
			return TRUE;
		}
	}
	
	return FALSE;
}



/*
 *	Insert a simple type into the data logger array. If the
 *	simple type is compatible with the one defined when the
 *	logger was initialised, then the operation will succeed
 *	unless there is no more memory left.
 *
 *	If the logging type is declared as complex, the input
 *	data will be converted into a "OPValue" wrapper before
 *	being stored.
 *
 *	If the types are incompatible, don't try to convert from
 *	one to another, even if this is possible (e.g. Card to
 *	Int or any other kind of conversion).
 */

Bool
OPDataLog::InsertCard32 (Card32 x)
{
	if (this->type == OP_TYPE_CARD) {
		return this->InsertAnyType (&x);
	} else if (this->type == OP_TYPE_OBJECT) {
		OPValueRef value = x;
		return this->InsertValue (value);
	}
	return FALSE;
}

Bool
OPDataLog::InsertInt32 (Int32 x)
{
	if (this->type == OP_TYPE_INT) {
		return this->InsertAnyType (&x);
	} else if (this->type == OP_TYPE_OBJECT) {
		OPValueRef value = x;
		return this->InsertValue (value);
	}
	return FALSE;
}

Bool
OPDataLog::InsertBool (Bool x)
{
	if (this->type == OP_TYPE_BOOL) {
		return this->InsertAnyType (&x);
	} else if (this->type == OP_TYPE_OBJECT) {
		OPValueRef value = x;
		return this->InsertValue (value);
	}
	return FALSE;
}

Bool
OPDataLog::InsertReal (Real x)
{
	if (this->type == OP_TYPE_REAL) {
		return this->InsertAnyType (&x);
	} else if (this->type == OP_TYPE_OBJECT) {
		OPValueRef value = x;
		return this->InsertValue (value);
	}
	return FALSE;
}


Bool
OPDataLog::GetNthCard32 (Count index, Card32Out x) const
{
	OP_METHOD (GetNthCard32, 2);
	OP_ARG (1, Count, index);
	OP_ARG (2, Card32Out, x);
	
	if (index < this->index) {
		
		if (this->is_cyclic) {
			index %= this->cyclic_size;
		}
		
		if ( (index < this->array_size)
		  && (this->is_simple)
		  && (this->type == OP_TYPE_CARD) ) {
			x = this->simple.card_array[index];
			return TRUE;
		}
	}
	
	return FALSE;
}

Bool
OPDataLog::GetNthInt32 (Count index, Int32Out x) const
{
	OP_METHOD (GetNthInt32, 2);
	OP_ARG (1, Count, index);
	OP_ARG (2, Int32Out, x);
	
	if (index < this->index) {
		
		if (this->is_cyclic) {
			index %= this->cyclic_size;
		}
		
		if ( (index < this->array_size)
		  && (this->is_simple)
		  && (this->type == OP_TYPE_INT) ) {
			x = this->simple.int_array[index];
			return TRUE;
		}
	}
	
	return FALSE;
}

Bool
OPDataLog::GetNthBool (Count index, BoolOut x) const
{
	OP_METHOD (GetNthBool, 2);
	OP_ARG (1, Count, index);
	OP_ARG (2, BoolOut, x);
	
	if (index < this->index) {
		
		if (this->is_cyclic) {
			index %= this->cyclic_size;
		}
		
		if ( (index < this->array_size)
		  && (this->is_simple)
		  && (this->type == OP_TYPE_BOOL) ) {
			x = this->simple.bool_array[index];
			return TRUE;
		}
	}
	
	return FALSE;
}

Bool
OPDataLog::GetNthReal (Count index, RealOut x) const
{
	OP_METHOD (GetNthReal, 2);
	OP_ARG (1, Count, index);
	OP_ARG (2, RealOut, x);
	
	if (index < this->index) {
		
		if (this->is_cyclic) {
			index %= this->cyclic_size;
		}
		
		if ( (index < this->array_size)
		  && (this->is_simple)
		  && (this->type == OP_TYPE_REAL) ) {
			x = this->simple.real_array[index];
			return TRUE;
		}
	}
	
	return FALSE;
}

Bool
OPDataLog::GetMultipleCard32 (Count index, Card32* x, Count num) const
{
	OP_METHOD (GetMultipleCard32, 3);
	OP_ARG (1, Count, index);
	OP_ARG (2, Card32*, x);
	OP_ARG (3, Count, num);
	
	if ( (index < this->index)
	  && (this->is_simple)
	  && (this->type == OP_TYPE_CARD) ) {
		
		if (this->is_cyclic) {
			index %= this->cyclic_size;
		}
		
		while (num--) {
			if (index >= this->array_size) {
				return FALSE;
			}
			
			*x++ = this->simple.card_array[index];
			index++;
			
			if ( (this->is_cyclic)
			  && (index == this->cyclic_size) ) {
				index = 0;
			}
		}
		
		return TRUE;
	}
	
	return FALSE;
}

Bool
OPDataLog::GetMultipleInt32 (Count index, Int32* x, Count num) const
{
	OP_METHOD (GetMultipleInt32, 3);
	OP_ARG (1, Count, index);
	OP_ARG (2, Int32*, x);
	OP_ARG (3, Count, num);
	
	if ( (index < this->index)
	  && (this->is_simple)
	  && (this->type == OP_TYPE_INT) ) {
		
		if (this->is_cyclic) {
			index %= this->cyclic_size;
		}
		
		while (num--) {
			if (index >= this->array_size) {
				return FALSE;
			}
			
			*x++ = this->simple.int_array[index];
			index++;
			
			if ( (this->is_cyclic)
			  && (index == this->cyclic_size) ) {
				index = 0;
			}
		}
		
		return TRUE;
	}
	
	return FALSE;
}

Bool
OPDataLog::GetMultipleBool (Count index, Bool* x, Count num) const
{
	OP_METHOD (GetMultipleBool, 3);
	OP_ARG (1, Count, index);
	OP_ARG (2, Bool*, x);
	OP_ARG (3, Count, num);
	
	if ( (index < this->index)
	  && (this->is_simple)
	  && (this->type == OP_TYPE_BOOL) ) {
		
		if (this->is_cyclic) {
			index %= this->cyclic_size;
		}
		
		while (num--) {
			if (index >= this->array_size) {
				return FALSE;
			}
			
			*x++ = this->simple.bool_array[index];
			index++;
			
			if ( (this->is_cyclic)
			  && (index == this->cyclic_size) ) {
				index = 0;
			}
		}
		
		return TRUE;
	}
	
	return FALSE;
}

Bool
OPDataLog::GetMultipleReal (Count index, Real* x, Count num) const
{
	OP_METHOD (GetMultipleReal, 3);
	OP_ARG (1, Count, index);
	OP_ARG (2, Real*, x);
	OP_ARG (3, Count, num);
	
	if ( (index < this->index)
	  && (this->is_simple)
	  && (this->type == OP_TYPE_REAL) ) {
		
		if (this->is_cyclic) {
			index %= this->cyclic_size;
		}
		
		while (num--) {
			if (index >= this->array_size) {
				return FALSE;
			}
			
			*x++ = this->simple.real_array[index];
			index++;
			
			if ( (this->is_cyclic)
			  && (index == this->cyclic_size) ) {
				index = 0;
			}
		}
		
		return TRUE;
	}
	
	return FALSE;
}




Count
OPDataLog::TotalCount () const
{
	return this->index;
}

Bool
OPDataLog::Get (Count index, OPObjectOut object)
{
	OPValueRef value;
	Bool result = this->GetNthValue (index, value);
	object = value;
	return result;
}

Bool
OPDataLog::SpecifyExtraction (OPArrayIn x)
{
	return FALSE;
}

