/*
 *	dataslot.cxx
 *
 *	This file implements the OPaC Data Slot class which is used to export
 *	data from an application to its user interface.
 *
 *	(C) Copyright 1994-1998, Pierre ARNAUD, OPaC bright ideas
 *		CH-1437 SUSCEVAZ, Switzerland. All rights reserved.
 */

/******************************************************************************

  This file is part of the OPaC class library ("OPaC").

  OPaC is distributed with NO WARRANTY OF ANY KIND.  No author or distributor
  accepts any responsibility for the consequences of using it, or for whether
  it serves any particular purpose or works at all, unless he or she says so
  in writing. Refer to the OPaC Free Public License (the "License") for full
  details.

  Every copy of OPaC must include a copy of the License, normally in a plain
  ASCII text file named PUBLIC. The License grants you the right to copy,
  modify and redistribute OPaC, but only under certain conditions described
  in the License. The copyright notice and this notice must be preserved on
  all copies.

******************************************************************************/

#include <OPaC/array.h>
#include <OPaC/database.h>
#include <OPaC/dataslot.h>
#include <OPaC/new.h>
#include <OPaC/string.h>
#include <OPaC/atom.h>

/****************************************************************/

OP_IMPLEMENTATION (dataslot.h);
OP_IMPLEMENT_CLASS (OPDataSlot);

/****************************************************************/

#define	FIELD_NAME		1		//	attribute # for name
#define	FIELD_DATA		2		//	attribute # for data
#define	FIELD_OBSERVER	3		//	attribute # for observer
#define	MAX_FIELD		3

#define	CONTROL_NAME	1		//	attribute # for name
#define	CONTROL_BODY	2		//	attribute # for body
#define	MAX_CONTROL		2

#define	X_NONE			0x0000
#define X_FIELD_ONLY	OP_BIT(0)
#define	X_CONTROL_ONLY	OP_BIT(1)
#define	X_ALL			(X_FIELD_ONLY|X_CONTROL_ONLY)

/****************************************************************/

/*
 *	Initialise an empty data slot.
 */

OPDataSlot::OPDataSlot ()
{
	OP_METHOD (OPDataSlot, 0);

	OPMemZone* zone = this->ReturnMemZone ();

	this->db_field         = OP_NEW_FROM_ZONE (zone, OPDataBase);
	this->db_field_model   = OP_NEW_FROM_ZONE (zone, OPEntityModel);
	this->db_control       = OP_NEW_FROM_ZONE (zone, OPDataBase);
	this->db_control_model = OP_NEW_FROM_ZONE (zone, OPEntityModel);

	this->db_field_model->Initialise (this->db_field, "DataSlot:Field", MAX_FIELD);
	this->db_field_model->AddAttribute ("FieldName");
	this->db_field_model->AddAttribute ("FieldData");
	this->db_field_model->AddAttribute ("FieldObserver");

	this->db_control_model->Initialise (this->db_control, "DataSlot:Control", MAX_CONTROL);
	this->db_control_model->AddAttribute ("ControllerName");
	this->db_control_model->AddAttribute ("ControllerBody");
	
	this->how_to_extract = X_ALL;
	this->name = "?";
}



/*
 *	Clean up a data slot.
 */

OPDataSlot::~OPDataSlot ()
{
	OP_METHOD (~OPDataSlot, 0);

	this->db_field   = 0;
	this->db_control = 0;
}



/*
 *	Store the data slot. This saves the data bases and the
 *	data they refer to. Be careful not to put temporaries
 *	into the data bases, or they will hang around when they
 *	get restored.
 */

void
OPDataSlot::StoreInstance (OPStorageIn storage)
{
	OP_METHOD (StoreInstance, 1);
	OP_ARG (1, OPStorageIn, storage);
	
	super::StoreInstance (storage);
	
	storage << this->db_field << this->db_control << this->how_to_extract;
	storage << OPStorageSync::mark;
}



/*
 *	Restore the data slot. This restores the data bases and
 *	the data they refer to.
 */

void
OPDataSlot::RestoreInstance (OPStorageIn storage)
{
	OP_METHOD (RestoreInstance, 1);
	OP_ARG (1, OPStorageIn, storage);
	
	super::RestoreInstance (storage);
	
	storage >> this->db_field >> this->db_control >> this->how_to_extract;
	storage >> OPStorageSync::mark;
}



/*
 *	Prepare the data slot for its destruction.
 */

void
OPDataSlot::Delete ()
{
	OP_METHOD (Delete, 0);
	
	if (this->db_control.IsValid ()) {
		this->db_control->RemoveAll ();
		this->db_control->Delete ();
	}

	if (this->db_control_model.IsValid ()) {
		this->db_control_model->Delete ();
	}
	
	if (this->db_field.IsValid ()) {
		this->db_field->RemoveAll ();
		this->db_field->Delete ();
	}

	if (this->db_field_model.IsValid ()) {
		this->db_field_model->Delete ();
	}

	this->db_field         = 0;
	this->db_field_model   = 0;
	this->db_control       = 0;
	this->db_control_model = 0;

	super::Delete ();
}



/*
 *	Return the total count of elements. This will change
 *	depending on the extraction mode.
 */

Count
OPDataSlot::TotalCount () const
{
	OP_METHOD (TotalCount, 0);
	
	Count count = 0;
	
	if ( (this->how_to_extract & X_FIELD_ONLY)
	  && (this->db_field.IsValid ()) ) {
		count += this->db_field->TotalCount ();
	}
	
	if ( (this->how_to_extract & X_CONTROL_ONLY)
	  && (this->db_control.IsValid ()) ) {
		count += this->db_control->TotalCount ();
	}
	
	return count;
}



/*
 *	Get the nth object.
 */

Bool
OPDataSlot::Get (Count index, OPObjectOut object)
{
	OP_METHOD (Get, 2);
	OP_ARG (1, Count, index);
	OP_ARG (2, OPObjectOut, object);
	
	Count count = 0;
	
	if ( (this->how_to_extract & X_FIELD_ONLY)
	  && (this->db_field.IsValid ()) ) {
		count = this->db_field->TotalCount ();
		if (index < count) {
			return this->db_field->Get (index, object);
		}
		index -= count;
	}
	
	if ( (this->how_to_extract & X_CONTROL_ONLY)
	  && (this->db_control.IsValid ()) ) {
		count = this->db_control->TotalCount ();
		if (index < count) {
			return this->db_control->Get (index, object);
		}
		index -= count;
	}
	
	return FALSE;
}



/*
 *	Define how the extraction will have to be handled.
 *
 *	You may specify either:
 *
 *	- ExtractNone, don't extract anything
 *	- ExtractField [sort, name], extract just the fields
 *	- ExtractControl [sort, name], extract just the controls
 *	- ExtractAll, sort1, name1, sort2, name2, extract both
 *
 *	[sort, name] are optional and can be used to specify how
 *	the extraction will be sorted (by name, object, observer
 *	or by name, body) and which field will be retrieved.
 */

Bool
OPDataSlot::SpecifyExtraction (OPArrayIn x)
{
	OP_METHOD (SpecifyExtraction, 1);
	OP_ARG (1, OPArrayIn, x);
	
	static OPAtomRef atom_x_none    = "ExtractNone";
	static OPAtomRef atom_x_field   = "ExtractField";
	static OPAtomRef atom_x_control = "ExtractControl";
	static OPAtomRef atom_x_all     = "ExtractAll";
	static OPAtomRef atom_x_attrib  = "ExtractAttribute";
	static OPAtomRef atom_f_name    = "FieldName";
	static OPAtomRef atom_c_name    = "ControllerName";
	
	Count  index = 0;

	OPAtomRef how = x[0];

	this->how_to_extract &= ~X_ALL;

	if ( (how == atom_x_field)
	  || (how == atom_x_all) ) {
		
		if (this->db_field.IsValid ()) {
			
			OPAtomArray spec;
			
			spec = OP_NEW (OPArray);
			spec->SetSize (3);
			
			if (x->TotalCount () >= 3) {
				spec[0] = atom_x_attrib;
				spec[1] = x[1];
				spec[2] = x[2];
				index  += 2;
			} else {
				spec[0] = atom_x_attrib;
				spec[1] = atom_f_name;
				spec[2] = atom_f_name;
			}
			
			this->db_field->SpecifyExtraction (spec);
		}
		
		this->how_to_extract |= X_FIELD_ONLY;
	}
	
	if ( (how == atom_x_control)
	  || (how == atom_x_all) ) {
		
		if (this->db_control.IsValid ()) {
			
			OPAtomArray spec;
			
			spec = OP_NEW (OPArray);
			spec->SetSize (3);
			
			if (x->TotalCount () >= (3 + index)) {
				spec[0] = atom_x_attrib;
				spec[1] = x[index+1];
				spec[2] = x[index+2];
				index  += 2;
			} else {
				spec[0] = atom_x_attrib;
				spec[1] = atom_c_name;
				spec[2] = atom_c_name;
			}
			
			this->db_control->SpecifyExtraction (spec);
		}
		
		this->how_to_extract |= X_CONTROL_ONLY;
	}
	
	return TRUE;
}


/*
 *	Each slot has a name which can be used to provide some
 *	user friendly information while linking properties
 *	together.
 */

void
OPDataSlot::SetName (OPAtomIn name)
{
	OP_METHOD (SetName, 1);
	OP_ARG (1, OPAtomIn, name);
	this->name = name;
}

Bool
OPDataSlot::GetName (OPAtomOut name)
{
	OP_METHOD (SetName, 1);
	OP_ARG (1, OPAtomIn, name);
	name = this->name;
	return TRUE;
}


/*
 *	Remember about the data field. If a data field is inserted
 *	twice, this will produce an error (NIY).
 */

void
OPDataSlot::AddField (OPAtomIn name, OPObjectInOut data)
{
	OP_METHOD (AddField, 2);
	OP_ARG (1, OPAtomIn, name);
	OP_ARG (2, OPObjectInOut, data);
	
	static OPAtomRef field_name = "FieldName";
	
	OPMemZone*  zone = 0;
	OPEntityRef slot;
	
	if (this->db_field->Find (field_name, name, slot)) {
		slot->SetAttribute (FIELD_DATA, data);
		return;
	}
	
	zone = this->ReturnMemZone ();
	slot = OP_NEW_FROM_ZONE (zone, OPEntity);
	
	slot->Initialise (this->db_field_model);
	slot->SetAttribute (FIELD_NAME, name);
	slot->SetAttribute (FIELD_DATA, data);

	this->db_field->Insert (slot);
}


/*
 *	Same as above, but an observer is associated to the data.
 *	If the data changes, the observer will be notified.
 */

void
OPDataSlot::AddField (OPAtomIn name, OPObjectInOut data, OPTargetIn observer)
{
	OP_METHOD (AddField, 3);
	OP_ARG (1, OPAtomIn, name);
	OP_ARG (2, OPObjectInOut, data);
	OP_ARG (3, OPTargetIn, observer);
	
	static OPAtomRef field_name = "FieldName";
	
	OPMemZone*  zone = 0;
	OPEntityRef slot;
	
	if (this->db_field->Find (field_name, name, slot)) {
		slot->SetAttribute (FIELD_DATA, data);
		slot->SetAttribute (FIELD_OBSERVER, observer);
		return;
	}
	
	zone = this->ReturnMemZone ();
	slot = OP_NEW_FROM_ZONE (zone, OPEntity);
	
	slot->Initialise (this->db_field_model);
	slot->SetAttribute (FIELD_NAME, name);
	slot->SetAttribute (FIELD_DATA, data);
	slot->SetAttribute (FIELD_OBSERVER, observer);

	this->db_field->Insert (slot);
}



/*
 *	Remember about the controller. (NIY)
 */

void
OPDataSlot::AddController (OPAtomIn name, int controller)
{
	OP_METHOD (AddController, 2);
	OP_ARG (1, OPAtomIn, name);
	OP_ARG (2, int, controller);

	//	NIY
}



/*
 *	Find the proper object stored in the named field. This
 *	returns TRUE on success.
 */

Bool
OPDataSlot::FindField (OPAtomIn name, OPObjectOut data)
{
	OP_METHOD (FindField, 2);
	OP_ARG (1, OPAtomIn, name);
	OP_ARG (2, OPObjectOut, data);
	
	OPTargetRef observer;
	
	return this->FindField (name, data, observer);
}



/*
 *	Find the proper object stored in the named field. This
 *	returns TRUE on success. If an observer has been linked
 *	to the field, return it too.
 */

Bool
OPDataSlot::FindField (OPAtomIn name, OPObjectOut data, OPTargetOut observer)
{
	OP_METHOD (FindField, 3);
	OP_ARG (1, OPAtomIn, name);
	OP_ARG (2, OPObjectOut, data);
	OP_ARG (3, OPTargetOut, observer);
	
	OPAtomRef   sort = "FieldName";
	Count       num  = this->db_field->TotalCount ();
	Count       pos  = 0;
	OPEntityRef entity;
	OPAtomRef   field_name;
	
	for (pos = 0; pos < num; pos++) {
		if ( (this->db_field->GetEntity (sort, pos, entity))
		  && (entity->GetAttribute (FIELD_NAME, field_name))
		  && (field_name == name) ) {
			
			entity->GetAttribute (FIELD_DATA, data);
			entity->GetAttribute (FIELD_OBSERVER, observer);
			
			return TRUE;
		}
	}
	
	return FALSE;
}



/*
 *	Find the proper controller. NIY.
 */

Bool
OPDataSlot::FindController (OPAtomIn name, int controller)
{
	OP_METHOD (FindController, 2);
	OP_ARG (1, OPAtomIn, name);
	OP_ARG (2, int, controller);
	
	//	NIY

	return FALSE;
}



/*
 *	Return the field or controller data bases. This will be
 *	used by the dialog in order to publish these informations.
 */

const OPDataBaseRef&
OPDataSlot::ReturnFieldBase () const
{
	OP_METHOD (ReturnFieldBase, 0);
	return this->db_field;
}

const OPDataBaseRef&
OPDataSlot::ReturnControllerBase () const
{
	OP_METHOD (ReturnControllerBase, 0);
	return this->db_control;
}

