/*
 *	dbprovider.cxx
 *
 *	The Data base Provider returns objects from the data base and
 *	uses its field names...
 *	
 *	(C) Copyright 1997-1998, Pierre ARNAUD, OPaC bright ideas
 *		CH-1437 SUSCEVAZ, Switzerland. All rights reserved.
 */

/******************************************************************************

  This file is part of the OPaC class library ("OPaC").

  OPaC is distributed with NO WARRANTY OF ANY KIND.  No author or distributor
  accepts any responsibility for the consequences of using it, or for whether
  it serves any particular purpose or works at all, unless he or she says so
  in writing. Refer to the OPaC Free Public License (the "License") for full
  details.

  Every copy of OPaC must include a copy of the License, normally in a plain
  ASCII text file named PUBLIC. The License grants you the right to copy,
  modify and redistribute OPaC, but only under certain conditions described
  in the License. The copyright notice and this notice must be preserved on
  all copies.

******************************************************************************/

#include <OPaC/database.h>
#include <OPaC/dataslot.h>
#include <OPaC/icon.h>
#include <OPaC/lookman.h>
#include <OPaC/objectpane.h>
#include <OPaC/values.h>

/****************************************************************/

OP_IMPLEMENT_CLASS (OPDBProvider);

/****************************************************************/

OPDBProvider::OPDBProvider ()
{
	OP_METHOD (OPDBProvider, 0);
	
	this->db_sort_name = "FieldName";
	this->db_data_name = "FieldData";
	this->db_text_name = "FieldName";
	
	this->db_large_icon_name = OPAtomNull;
	this->db_small_icon_name = OPAtomNull;
	this->db_mark_icon_name  = OPAtomNull;
}

OPDBProvider::~OPDBProvider ()
{
	OP_METHOD (~OPDBProvider, 0);
}


void
OPDBProvider::StoreInstance (OPStorageIn storage)
{
	OP_METHOD (StoreInstance, 1);
	OP_ARG (1, OPStorageIn, storage);
	
	super::StoreInstance (storage);
	
	storage << this->db_sort_name << this->db_data_name << this->db_text_name;
	storage << this->db_large_icon_name << this->db_small_icon_name;
	storage << this->db_mark_icon_name;
	storage << OPStorageSync::mark;
}

void
OPDBProvider::RestoreInstance (OPStorageIn storage)
{
	OP_METHOD (RestoreInstance, 1);
	OP_ARG (1, OPStorageIn, storage);
	
	super::RestoreInstance (storage);
	
	storage >> this->db_sort_name >> this->db_data_name >> this->db_text_name;
	storage >> this->db_large_icon_name >> this->db_small_icon_name;
	storage >> this->db_mark_icon_name;
	storage >> OPStorageSync::mark;
}


/*
 *	Find the icon code from an object (which may either
 *	be a name, a number or an icon view
 */

OPLookIcon
OPDBProvider::MapObjectToIcon (OPObjectIn object) const
{
	OP_METHOD (MapObjectToIcon, 1);
	OP_ARG (1, OPObjectIn, object);
	
	if (object.IsValid ()) {
		
		if (object->IsKindOf (OP_RUNTIME (OPText))) {
			
			OPTextRef name = object;
			return OPLookMan::Current ()->FindIcon (name);
			
		} else if (object->IsKindOf (OP_RUNTIME (OPValue))) {
			
			OPValueRef   value = object;
			OPSimpleType type  = value->ReturnType ();
			switch (type) {
				case OP_TYPE_INT:
				case OP_TYPE_CARD:
				case OP_TYPE_ENUM:
					return (OPLookIcon) value->ReturnInt32 ();
			}
			
		} else if (object->IsKindOf (OP_RUNTIME (OPIcon))) {
			
			OPIconRef icon = object;
			return icon->ReturnNormalIcon ();
		}
	}
	
	return OP_ICON_NONE;
}


/*
 *	Define the data base names for extracting the sorting
 *	index, the object, the name, etc...
 */

Bool
OPDBProvider::SetDBNameForSorting (OPAtomIn name)
{
	OP_METHOD (SetDBNameForSorting, 1);
	OP_ARG (1, OPAtomIn, name);
	
	this->db_sort_name = name;
	
	return TRUE;
}

Bool
OPDBProvider::SetDBNameForObject (OPAtomIn name)
{
	OP_METHOD (SetDBNameForObject, 1);
	OP_ARG (1, OPAtomIn, name);
	
	this->db_data_name = name;
	
	return TRUE;
}

Bool
OPDBProvider::SetDBNameForText (OPAtomIn text)
{
	OP_METHOD (SetDBNameForText, 1);
	OP_ARG (1, OPAtomIn, text);
	
	this->db_text_name = text;
	
	return TRUE;
}

Bool
OPDBProvider::SetDBNameForLargeIcon (OPAtomIn name)
{
	OP_METHOD (SetDBNameForLargeIcon, 1);
	OP_ARG (1, OPAtomIn, name);
	
	this->db_large_icon_name = name;
	
	return TRUE;
}

Bool
OPDBProvider::SetDBNameForSmallIcon (OPAtomIn name)
{
	OP_METHOD (SetDBNameForSmallIcon, 1);
	OP_ARG (1, OPAtomIn, name);
	
	this->db_small_icon_name = name;
	
	return TRUE;
}

Bool
OPDBProvider::SetDBNameForMarkIcon (OPAtomIn name)
{
	OP_METHOD (SetDBNameForMarkIcon, 1);
	OP_ARG (1, OPAtomIn, name);
	
	this->db_mark_icon_name = name;
	
	return TRUE;
}

Bool
OPDBProvider::GetObject (Count index, OPObjectOut object)
{
	OP_METHOD (GetObject, 2);
	OP_ARG (1, Count, index);
	OP_ARG (2, OPObjectOut, object);
	
	if (this->db_data_name.IsValid ()) {
		OPDataBaseRef db = this->container;
		return db->GetAttribute (this->db_sort_name, index, this->db_data_name, object);
	}
	
	return super::GetObject (index, object);
}

Bool
OPDBProvider::GetLargeIcon (Count index, OPLookIconOut icon)
{
	OP_METHOD (GetLargeIcon, 2);
	OP_ARG (1, Count, index);
	OP_ARG (2, OPLookIconOut, icon);
	
	if (this->db_large_icon_name.IsValid ()) {
		
		OPDataBaseRef db = this->container;
		OPObjectRef object;
		
		db->GetAttribute (this->db_sort_name, index, this->db_large_icon_name, object);
		icon = this->MapObjectToIcon (object);
		
		return TRUE;
	}
	
	return super::GetLargeIcon (index, icon);
}

Bool
OPDBProvider::GetSmallIcon (Count index, OPLookIconOut icon)
{
	OP_METHOD (GetSmallIcon, 2);
	OP_ARG (1, Count, index);
	OP_ARG (2, OPLookIconOut, icon);
	
	if (this->db_small_icon_name.IsValid ()) {
		
		OPDataBaseRef db = this->container;
		OPObjectRef object;
		
		db->GetAttribute (this->db_sort_name, index, this->db_small_icon_name, object);
		icon = this->MapObjectToIcon (object);
		
		return TRUE;
	}
	
	return super::GetSmallIcon (index, icon);
}

Bool
OPDBProvider::GetText (Count index, OPStringOut string)
{
	OP_METHOD (GetText, 2);
	OP_ARG (1, Count, index);
	OP_ARG (2, OPStringOut, string);
	
	if (this->db_text_name.IsValid ()) {
		
		OPDataBaseRef db = this->container;
		OPTextRef     text;
		
		if (db->GetAttribute (this->db_sort_name, index, this->db_text_name, text)) {
			string = text->ReturnString ();
			return string.IsValid ();
		}
		
		return FALSE;
	}
	
	return super::GetText (index, string);
}

Bool
OPDBProvider::GetMarkIcon (Count index, OPLookIconOut icon)
{
	OP_METHOD (GetMarkIcon, 2);
	OP_ARG (1, Count, index);
	OP_ARG (2, OPLookIconOut, icon);
	
	if (this->db_mark_icon_name.IsValid ()) {
		
		OPDataBaseRef db = this->container;
		OPObjectRef object;
		
		db->GetAttribute (this->db_sort_name, index, this->db_mark_icon_name, object);
		icon = this->MapObjectToIcon (object);
		
		return TRUE;
	}
	
	return super::GetMarkIcon (index, icon);
}

