/*
 *	except.cxx
 *
 *	This file implements the C++ base exception class and the
 *	associated methods.
 *
 *	(C) Copyright 1994-1997, Pierre ARNAUD, OPaC bright ideas
 *		CH-1437 SUSCEVAZ, Switzerland. All rights reserved.
 */

/******************************************************************************

  This file is part of the OPaC class library ("OPaC").

  OPaC is distributed with NO WARRANTY OF ANY KIND.  No author or distributor
  accepts any responsibility for the consequences of using it, or for whether
  it serves any particular purpose or works at all, unless he or she says so
  in writing. Refer to the OPaC Free Public License (the "License") for full
  details.

  Every copy of OPaC must include a copy of the License, normally in a plain
  ASCII text file named PUBLIC. The License grants you the right to copy,
  modify and redistribute OPaC, but only under certain conditions described
  in the License. The copyright notice and this notice must be preserved on
  all copies.

******************************************************************************/

#include <OPaC/except.h>
#include <OPaC/new.h>

OP_IMPLEMENTATION (except.h);

/*
 *	This file implements following classes :
 *
 *	  -	OPaC_ExcLink, stack class which records the "try"
 *		levels.
 *	
 *	  -	OPException, generic exception class.
 */

#define	MAX_EXC_SIZE	200

/*
 *	The OPaC exception mechanisms need to maintain a stack
 *	of TRY contexts, that's why there are some global, per
 *	thread, variables.
 *
 *	CAUTION : This will need some modification the day we
 *			  get multi-threaded OPaC !
 */

OPaC_ExcLink* OPaC_ExcLink::top_link   = 0;
static Card8  exc_buffer[MAX_EXC_SIZE] = { 0 /*...*/ };

/****************************************************************/

/*
 *	Prepare for a TRY block. This adds a new context in
 *	the try stack.
 */
	
OPaC_ExcLink::OPaC_ExcLink ()
{
	OP_MEMBER (OPaC_ExcLink, 0);
	
	this->current_exc_inst = 0;
	this->current_exc_name = "";
	this->prev             = OPaC_ExcLink::top_link;
	
	OPaC_ExcLink::top_link = this;
}



/*
 *	Either `Exception' or the destructor will be invoked.
 *	If it is the destructor, then nothing has to be cleaned
 *	up and the previous state can just be popped back.
 */

OPaC_ExcLink::~OPaC_ExcLink ()
{
	OP_MEMBER (~OPaC_ExcLink, 0);
	
	this->top_link = this->prev;
}



/*
 *	Look if the specified exception matches the one that
 *	has been thrown previously.
 */

Bool
OPaC_ExcLink::Matching (const LatinChar* name)
{
	OP_MEMBER (Matching, 1);
	OP_ARG (1, const LatinChar*, name);
	
	//	If there is no exception, then there would be no reason to catch
	//	anything... This would mean OPaC didn't compile well on this target
	//	machine/system.
	
	OP_ASSERT (this->current_exc_inst);
	
	const LatinChar* ptr = this->current_exc_name;
	
	while (*ptr) {
		if ( (*ptr == ' ')
		  || (*ptr == '(')
		  || (*ptr == ';') )
			break;				
		
		if (*ptr != *name)
			return FALSE;
		
		ptr++;
		name++;
	}
	
	return (*name == 0) ? TRUE : FALSE;
}



/*
 *	A matching CATCH block has been found; return the
 *	exception instance after having restored the previous
 *	state.
 */

OPException*
OPaC_ExcLink::Exception ()
{
	OP_MEMBER (Exception, 0);
	
	this->top_link = this->prev;
	
	return this->current_exc_inst;
}



/*
 *	We didn't find any matching CATCH block, so we have to
 *	propagate the exception to the next level, after having
 *	restored the previous state (done by `Exception').
 */

void
OPaC_ExcLink::EndCatch ()
{
	OP_MEMBER (EndCatch, 0);
	
	OPException*     exc_inst = this->Exception ();
	const LatinChar* exc_name = this->current_exc_name;
	
	OPaC_ExcLink::ThrowException ( exc_inst, exc_name,
								   exc_inst->throw_line,
								   exc_inst->throw_file );
}



/*
 *	Throws an exception, which will be handled by the
 *	latest CATCH block encountered. First clean up every
 *	temporary object found in the current try context.
 */

void
OPaC_ExcLink::ThrowException (OPException* exception, const LatinChar* name,
							  Count throw_line, const LatinChar* throw_file)
{
	OP_MEMBER (ThrowException, 4);
	OP_ARG (1, OPException*, exception);
	OP_ARG (2, const LatinChar*, name);
	OP_ARG (3, Count, throw_line);
	OP_ARG (4, const LatinChar*, throw_file);
	
	OP_ASSERT (OPaC_ExcLink::top_link);
	OP_ASSERT (exception);
	OP_ASSERT (name);
	
	OPaC_ExcLink* link = OPaC_ExcLink::top_link;
	
	top_link->current_exc_inst = exception;
	top_link->current_exc_name = name;
	
	exception->throw_line = throw_line;
	exception->throw_file = throw_file;
	
	longjmp (link->caller, 1);
}

/****************************************************************/

/*
 *	Allocate memory for an exception. This will always use
 *	the same static buffer, since only one exception at a
 *	time can be tolerated.
 */

void*
OPException::operator new (size_t size)
{
	OP_OPERATOR (new, 1);
	OP_ARG (1, size_t, size);
	OP_ASSERT (size <= MAX_EXC_SIZE);
	
	return exc_buffer;
}



/*
 *	Return the source line number where the throw occurred.
 */

Count
OPException::GetThrowSrcLine () const
{
	OP_MEMBER (GetThrowSrcLine, 0);
	return this->throw_line;
}



/*
 *	Return the source file name where the throw occurred.
 */

const LatinChar*
OPException::GetThrowSrcFile () const
{
	OP_MEMBER (GetThrowSrcFile, 0);
	return this->throw_file;
}



