/*
 *	grailcomp.cxx
 *
 *	This file implements the Grail Computation functions.
 *
 *	(C) Copyright 1994-1997, Pierre ARNAUD, OPaC bright ideas
 *		CH-1437 SUSCEVAZ, Switzerland. All rights reserved.
 */

/******************************************************************************

  This file is part of the OPaC class library ("OPaC").

  OPaC is distributed with NO WARRANTY OF ANY KIND.  No author or distributor
  accepts any responsibility for the consequences of using it, or for whether
  it serves any particular purpose or works at all, unless he or she says so
  in writing. Refer to the OPaC Free Public License (the "License") for full
  details.

  Every copy of OPaC must include a copy of the License, normally in a plain
  ASCII text file named PUBLIC. The License grants you the right to copy,
  modify and redistribute OPaC, but only under certain conditions described
  in the License. The copyright notice and this notice must be preserved on
  all copies.

******************************************************************************/

#include <OPaC/grail.h>
#include <OPaC/new.h>

/****************************************************************/

OP_MORE_IMPLEMENTATION (OPGrailPort);

/****************************************************************/

/*
 *	Compute the Y co-ordinates of a line. This requires that :
 *	
 *	- dx  >  0,   the line is drawn from left to right
 *	- dx >= |dy|, the angle lies between -45 and +45 degrees.
 *
 *	Both arrays will be filled with dx+1 values. The y_error
 *	array (if any) will be filled with errors ranging from
 *	0 to 255, with 0 specifying an exact match. This can be
 *	used for anti-aliasing.
 *
 *	The error parameter can be used to specify an initial
 *	vertical error (useful only in the range 0 to 0.996).
 */

void
OPGrailPort::ComputeLineDXDots (Card32 x, Card32 y, Card32 dx, Int32 dy,
								Card32* y_array, Card8* y_error, Real error)
{
	if (error < 0) {
		Card32 num = (Card32)(-error);
		error += (Real)(num+1);
		y     -= num;
	} else if (error >= 1.0) {
		Card32 num = (Card32)(error);
		error -= (Real)(num);
		y     += num;
	}
	
	Real m = (Real)(dy) / (Real)(dx);
	Real e = error;
	
	if (m >= 0) {
		
		while (dx--) {
			
			if (y_error) *y_error++ = (Card8)(e * 255.99);
			if (y_array) *y_array++ = y;
			
			e += m;
			
			if (e >= 0.9999) {
				e -= 1.0;
				y++;
			}
		}
		
		if (y_error) *y_error++ = (Card8)(e * 255.99);
		if (y_array) *y_array++ = y;
		
	} else {
		
		while (dx--) {
			
			if (y_error) *y_error++ = (Card8)(e * 255.99);
			if (y_array) *y_array++ = y;
			
			e += m;
			
			if (e < 0.0) {
				e += 1.0;
				y--;
			}
		}
		
		if (y_error) *y_error++ = (Card8)(e * 255.99);
		if (y_array) *y_array++ = y;
	}
}

/*
 *	Compute the X co-ordinates of a line. This requires that :
 *	
 *	- dy  >  0,   the line is drawn from top to bottom
 *  - dy >= |dx|, the angle lies between +45 and +135 degrees.
 *
 *	Same remarks as for ComputeLineDXDots.
 */

void
OPGrailPort::ComputeLineDYDots (Card32 x, Card32 y, Int32 dx, Card32 dy,
								Card32* x_array, Card8* x_error, Real error)
{
	Card32 end_x = x + dx;
	
	if (error < 0) {
		Card32 num = (Card32)(-error);
		error += (Real)(num+1);
		x     -= num;
	} else if (error >= 1.0) {
		Card32 num = (Card32)(error);
		error -= (Real)(num);
		x     += num;
	}
	
	Real m = (Real)(dx) / (Real)(dy);
	Real e = error;
	
	if (m >= 0) {
		
		while (dy--) {
			
			if (x_error) *x_error++ = (Card8)(e * 255.99);
			if (x_array) *x_array++ = x;
			
			e += m;
			
			if (e >= 0.9999) {
				e -= 1.0;
				x++;
			}
		}
		
		if (x_error) *x_error++ = (Card8)(e * 255.99);
		if (x_array) *x_array++ = x;
		
	} else {
		
		while (dy--) {
			
			if (x_error) *x_error++ = (Card8)(e * 255.99);
			if (x_array) *x_array++ = x;
			
			e += m;
			
			if (e < 0.0) {
				e += 1.0;
				x--;
			}
		}
		
		if (x_error) *x_error++ = (Card8)(e * 255.99);
		if (x_array) *x_array++ = x;
	}
}


