/*
 *	loadgif.cxx
 *
 *	Originally gif2ras.c. It has the following copyright :
 *
 *	Copyright (c) 1988, 1989 by Patrick J. Naughton
 *
 *	Author: Patrick J. Naughton
 *	naughton@wind.sun.com
 *	
 *	Permission to use, copy, modify, and distribute this software and its
 *	documentation for any purpose and without fee is hereby granted,
 *	provided that the above copyright notice appear in all copies and that
 *	both that copyright notice and this permission notice appear in
 *	supporting documentation.
 *
 *	This file is provided AS IS with no warranties of any kind.	The author
 *	shall have no liability with respect to the infringement of copyrights,
 *	trade secrets or any patents by this file or any part thereof.  In no
 *	event will the author be liable for any lost revenue or profits or
 *	other special, indirect and consequential damages.
 *
 *	CAUTION :	This file has been heavily modified by Pierre ARNAUD
 *				(Pierre.Arnaud@iname.com) in order to fit into the OPaC
 *				project.
 *
 *	The modifications are copyrighted by Pierre ARNAUD :
 *
 *	(C) Copyright 1994-1998, Pierre ARNAUD, OPaC bright ideas
 *		CH-1437 SUSCEVAZ, Switzerland. All rights reserved.
 */

/******************************************************************************

  This file is part of the OPaC class library ("OPaC").

  OPaC is distributed with NO WARRANTY OF ANY KIND.  No author or distributor
  accepts any responsibility for the consequences of using it, or for whether
  it serves any particular purpose or works at all, unless he or she says so
  in writing. Refer to the OPaC Free Public License (the "License") for full
  details.

  Every copy of OPaC must include a copy of the License, normally in a plain
  ASCII text file named PUBLIC. The License grants you the right to copy,
  modify and redistribute OPaC, but only under certain conditions described
  in the License. The copyright notice and this notice must be preserved on
  all copies.

******************************************************************************/

#include <OPaC/object.h>
#include <OPaC/stream.h>
#include <OPaC/grail.h>
#include <OPaC/new.h>

#include <string.h>

/****************************************************************/

#define NEXTBYTE (*ptr++)
#define EXTENSION	  0x21
#define IMAGESEP	  0x2c
#define TRAILER		  0x3b
#define INTERLACEMASK 0x40
#define COLORMAPMASK  0x80

/****************************************************************/

struct GIFloader {
	
	Int32	BitOffset;					/* Bit Offset of next code */
	Int32	XC;
	Int32	YC;							/* Output X and Y coords of current pixel */
	Int32	Pass;						/* Used by output routine if interlaced pic */
	Int32	OutCount;					/* Decompressor output 'stack count' */
	Int32	Width;
	Int32	Height;						/* image dimensions */
	Int32	BitsPerPixel;				/* Bits per pixel, read from GIF header */
	Int32	ColorMapSize;				/* number of colors */
	Int32	CodeSize;					/* Code size, read from GIF header */
	Int32	InitCodeSize;				/* Starting code size, used during Clear */
	Int32	Code;						/* Value returned by ReadCode */
	Int32	MaxCode;					/* limiting value for current code size */
	Int32	ClearCode;					/* GIF clear code */
	Int32	EOFCode;					/* GIF end-of-information code */
	Int32	CurCode;
	Int32	OldCode;
	Int32	InCode;						/* Decompressor variables */
	Int32	FirstFree;					/* First free code, generated per GIF spec */
	Int32	FreeCode;					/* Decompressor,next free slot in hash table */
	Int32	FinChar;					/* Decompressor variable */
	Int32	BitMask;					/* AND mask for data size */
	Int32	ReadMask;					/* Code AND mask for current code size */
	Int32	Misc;						/* miscellaneous bits (interlace, local cmap)*/
	
	Bool	Interlace;
	Bool	HasColormap;
	
	Card8*	pic;					/* The final image */
	Card8*	RawGIF;					/* The heap array to hold it, raw */
	Card8*	Raster;					/* The raster data stream, unblocked */
	Card8*	iPtr;					/* The DoInterlace function uses its own pointer */
	
	Int32	oldYC;
	Int32	Prefix[4096];			/* The hash table used by the decompressor */
	Int32	Suffix[4096];
	Int32	OutCode[4097];			/* An output array used by the decompressor */
	Bool	gif89;
	Size	filesize;
	
	Card8	r[256];
	Card8	g[256];
	Card8	b[256];
	
	OPaC_FileDesc fp;
};

/****************************************************************/

static const char* const id87 = "GIF87a";
static const char* const id89 = "GIF89a";

/****************************************************************/

static int EGApalette[16][3] = {
  {0,0,0},		 {0,0,128},		{0,128,0},	   {0,128,128}, 
  {128,0,0},	 {128,0,128},	{128,128,0},   {200,200,200},
  {100,100,100}, {100,100,255}, {100,255,100}, {100,255,255},
  {255,100,100}, {255,100,255}, {255,255,100}, {255,255,255} };
  
/****************************************************************/

static Int32 ReadCode (GIFloader&);
static void DoInterlace (GIFloader&, Card8);
static OPGrailPixmap* GifError (GIFloader&);

/****************************************************************/

/*
 *	Load a Graphics Interchange File (GIF) as defined by the
 *	CompuServe community.
 *
 *	This returns a pixelmap containing the picture, with the
 *	same depth as the specified window.
 */

OPGrailPixmapRef
OPaC_LoadGIF (OPStringIn fname, OPMemZone* zone)
{
	OP_FUNCTION (OPaC_LoadGIF, 2);
	OP_ARG (1, OPStringIn, fname);
	OP_ARG (2, OPMemZone*, zone);
	
	GIFloader* gifptr = OP_ALLOC (GIFloader);
	
	if (gifptr == 0) {
		return 0;
	}
	
	if (zone == 0) {
		zone = OPaC_data_zone;
	}
	
	GIFloader& gif = *gifptr;
	
	Card8  ch, ch1;
	Card8* ptr;
	Card8* ptr1;
	Card8* picptr;
	
	Int32 i, j, a, t;
	Int32 block     = 0;
	Int32 npixels   = 0;
	Int32 maxpixels = 0;
	
	//	Initialise variables
	
	ptr = (Card8*)(&gif);
	i   = sizeof (gif);
	t   = 0;
	
	while (i--) {
		*ptr++ = 0;
	}
	
	gif.iPtr  = 0;
	gif.oldYC = -1;
	
	gif.fp = OPaC_OpenFileRead (fname);
	
	if (gif.fp == 0) {
		return 0;
	}
	
	//	Find the size of the file */
	
	OPaC_SeekFileAbs (gif.fp, -1);
	gif.filesize = OPaC_TellFilePos (gif.fp);
	OPaC_SeekFileAbs (gif.fp, 0);
	
	//	The +256's are so we can read truncated GIF files without fear of segmentation violation
	
	ptr = gif.RawGIF = (Card8 *) OP_ALLOC_ARRAY (char, gif.filesize + 256);
	if (ptr == 0) {
		return GifError (gif);
	}
	
	gif.Raster = (Card8 *) OP_ALLOC_ARRAY (char, gif.filesize + 256);
	if (gif.Raster == 0) {
		return GifError (gif);
	}
	
	if (OPaC_ReadFile (gif.fp, ptr, gif.filesize) != gif.filesize) {
		return GifError (gif);
	}
	
	if (strncmp ((char *) ptr, id87, 6) == 0) {
		gif.gif89 = FALSE;
	} else if (strncmp ((char *) ptr, id89, 6) == 0) {
		gif.gif89 = TRUE;
	} else {
		return GifError (gif);
	}
	
	ptr += 10;
	
	//	Get variables from the GIF screen descriptor
	
	ch = NEXTBYTE;
	gif.HasColormap = (ch & COLORMAPMASK) ? TRUE : FALSE;
	
	gif.BitsPerPixel = (ch & 7) + 1;
	gif.ColorMapSize = 1 << gif.BitsPerPixel;
	gif.BitMask      = gif.ColorMapSize - 1;
	
	ptr += 2;
	
	//	Read in global colormap.
	
	if (gif.HasColormap) {
		for (i = 0; i < gif.ColorMapSize; i++) {
			gif.r[i] = NEXTBYTE;
			gif.g[i] = NEXTBYTE;
			gif.b[i] = NEXTBYTE;
		}
	} else {
		//	No colormap in GIF file => put std EGA palette (repeated 16 times)
		//	into colormap, for lack of anything better to do !
		
		for (i = 0; i < 256; i++) {
			gif.r[i] = EGApalette[i&15][0];
			gif.g[i] = EGApalette[i&15][1];
			gif.b[i] = EGApalette[i&15][2];
		}
	}
	
	//	Possible things at this point are :
	//
	//	- an application extension block
	//	- a comment extension block
	//	- an (optional) graphic control extension block followed by either
	//	  an image or a plaintext extension

	for (;;) {
		
		block = NEXTBYTE;
		
		if (block == EXTENSION) {		//	parse extension blocks
			
			int i, blocksize;
			int fn = NEXTBYTE;
			
			if (fn == 'R') {			//	skip GIF87 aspect extension
				
				blocksize = NEXTBYTE;
				ptr += blocksize;
				
			} else if (fn == 0xFE) {	//	skip comment extension
				
				int sbsize;
				
				do {
					sbsize = NEXTBYTE;
					ptr += sbsize;
				} while (sbsize);
				
			} else if (fn == 0x01) {	//	skip plain text extension
				
				int sbsize;
				int tgLeft, tgTop, tgWidth, tgHeight, cWidth, cHeight, fg, bg;
				
				sbsize	 = NEXTBYTE;
				tgLeft	 = NEXTBYTE;  tgLeft   += (NEXTBYTE) << 8;
				tgTop	 = NEXTBYTE;  tgTop	   += (NEXTBYTE) << 8;
				tgWidth	 = NEXTBYTE;  tgWidth  += (NEXTBYTE) << 8;
				tgHeight = NEXTBYTE;  tgHeight += (NEXTBYTE) << 8;
				cWidth	 = NEXTBYTE;
				cHeight	 = NEXTBYTE;
				fg		 = NEXTBYTE;
				bg		 = NEXTBYTE;
				
				for (i = 12 ; i < sbsize; i++) {
					(void)(NEXTBYTE);
				}
				
				do {
					sbsize = NEXTBYTE;
					ptr += sbsize;
				} while (sbsize);
				
			} else if (fn == 0xF9) {	//	graphic control extension
				
				int sbsize;
				
				do {
					sbsize = NEXTBYTE;
					ptr += sbsize;
				} while (sbsize);
				
			} else {					//	unknown extension
				
				int sbsize;
				
				do {
					sbsize = NEXTBYTE;
					ptr += sbsize;
				} while (sbsize);
				
			}
		
		} else if (block == IMAGESEP) {	//	Read an image
			
			break;
			
		} else if (block == TRAILER) {
			
			return GifError (gif);
			
		} else {
			
			return GifError (gif);
		}
	}
	
	//	Read in values from the image descriptor
	
	ptr += 4;
	
	ch = NEXTBYTE;	gif.Width  = ch + 0x100 * NEXTBYTE;
	ch = NEXTBYTE;	gif.Height = ch + 0x100 * NEXTBYTE;
	
	gif.Misc      = NEXTBYTE;
	gif.Interlace = (gif.Misc & INTERLACEMASK) ? TRUE : FALSE;
	
	if (gif.Misc & 0x80) {
		for (i = 0; i < (1 << ((gif.Misc & 0x07) + 1)); i++) {
			gif.r[i] = NEXTBYTE;
			gif.g[i] = NEXTBYTE;
			gif.b[i] = NEXTBYTE;
		}
	}
	
	//	Start reading the raster data. First we get the intial code size and
	//	compute decompressor constant values, based on this code size.
	
	gif.CodeSize  = NEXTBYTE;
	gif.ClearCode = (1 << gif.CodeSize);
	gif.EOFCode   = gif.ClearCode + 1;
	gif.FreeCode  = gif.FirstFree = gif.ClearCode + 2;
	
	//	The GIF spec has it that the code size used to compute the above values is the
	//	code size given in the file, but the code size used in compression/decompression
	//	is the code size given in the file plus one -- thus the ++.
	
	gif.CodeSize++;
	gif.InitCodeSize = gif.CodeSize;
	gif.MaxCode      = (1 << gif.CodeSize);
	gif.ReadMask     = gif.MaxCode - 1;
	
	//	UNBLOCK :
	//	
	//	Read the raster data. Here we just transpose it from the GIF array to the Raster
	//	array, turning it from a series of blocks into one long data stream, which makes
	//	life much easier for "ReadCode".
	
	t = (long)(ptr - gif.RawGIF);
	ptr1 = gif.Raster;
	
	do {
		
		ch = ch1 = NEXTBYTE;
		
		while (ch--) {
			*ptr1++ = NEXTBYTE;
		}
		
		if ((Size)(ptr - gif.RawGIF) > gif.filesize) {
			break;
		}
		
	} while (ch1);
	
	OP_FREE (gif.RawGIF);	gif.RawGIF = NULL;	//	We're done with the raw data now
	
	maxpixels = gif.Width * gif.Height;
	
	picptr = gif.pic = (Card8 *) OP_ALLOC_ARRAY (char, maxpixels);
	
	if (gif.pic == 0) return GifError (gif);
	
	//	Decompress the file, continuing until you see the GIF EOF code.
	//	One obvious enhancement is to add checking for corrupt files here.
	
	gif.Code = ReadCode (gif);
	
	while (gif.Code != gif.EOFCode) {
		
		//	Clear code sets everything back to its initial value, then reads the
		//	immediately subsequent code as uncompressed data.
		
		if (gif.Code == gif.ClearCode) {
			gif.CodeSize = gif.InitCodeSize;
			gif.MaxCode  = (1 << gif.CodeSize);
			gif.ReadMask = gif.MaxCode - 1;
			gif.FreeCode = gif.FirstFree;
			gif.Code     = ReadCode (gif);
			gif.CurCode  = gif.OldCode = gif.Code;
			gif.FinChar  = gif.CurCode & gif.BitMask;
			
			if (gif.Interlace == FALSE)
				*picptr++ = (Card8)(gif.FinChar);
			else
				DoInterlace (gif, (Card8)(gif.FinChar));
			
			npixels++;
		
		} else {
			
			//	If not a clear code, must be data : save same as CurCode and InCode.
			//	If we're at maxcode and didn't get a clear, stop loading !
			
			if (gif.FreeCode >= 4096) break;
			
			gif.CurCode = gif.InCode = gif.Code;
			
			//	If greater or equal to FreeCode, not in the hash table yet; repeat
			//	the last character decoded.
			
			if (gif.CurCode >= gif.FreeCode) {
				
				gif.CurCode = gif.OldCode;
				if (gif.OutCount > 4096) break;
				gif.OutCode[gif.OutCount++] = gif.FinChar;
			}
			
			//	Unless this code is raw data, pursue the chain pointed to by CurCode
			//	through the hash table to its end; each code in the chain puts its
			//	associated output code on the output queue.
			
			while (gif.CurCode > gif.BitMask) {
				
				if (gif.OutCount > 4096) break;
				gif.OutCode[gif.OutCount++] = gif.Suffix[gif.CurCode];
				gif.CurCode = gif.Prefix[gif.CurCode];
			}
			
			if (gif.OutCount > 4096) break;
			
			//	The last code in the chain is treated as raw data.
			
			gif.FinChar = gif.CurCode & gif.BitMask;
			gif.OutCode[gif.OutCount++] = gif.FinChar;
			
			//	Now we put the data out to the Output routine. It's been stacked
			//	LIFO, so deal with it that way...
			
			//	Safety thing : prevent exceeding range of 'pic'
			
			if (npixels + gif.OutCount > maxpixels)
				gif.OutCount = maxpixels - npixels;
			
			npixels += gif.OutCount;
			
			if (gif.Interlace == FALSE) {
				for (i = gif.OutCount - 1; i >= 0; i--)
					*picptr++ = (Card8)(gif.OutCode[i]);
			} else {
				for (i = gif.OutCount - 1; i >= 0; i--)
					DoInterlace(gif, (Card8)(gif.OutCode[i]));
			}
			
			gif.OutCount = 0;
			
			//	Build the hash table on-the-fly. No table is stored in the file.
			
			gif.Prefix[gif.FreeCode] = gif.OldCode;
			gif.Suffix[gif.FreeCode] = gif.FinChar;
			gif.OldCode = gif.InCode;
			
			//	Point to the next slot in the table. If we exceed the current MaxCode
			//	value, increment the code size unless it's already 12. If it is, do
			//	nothing : the next code decompressed better be CLEAR !
			
			gif.FreeCode++;
			
			if (gif.FreeCode >= gif.MaxCode) {
				if (gif.CodeSize < 12) {
					gif.CodeSize++;
					gif.MaxCode *= 2;
					gif.ReadMask = (1 << gif.CodeSize) - 1;
				}
			}
		}
		
		t++;
		if (t == 1024) {
			t = 0;
		}
		
		gif.Code = ReadCode (gif);
		if (npixels >= maxpixels) break;
	}
	
	OP_FREE (gif.Raster);
	gif.Raster = NULL;
	
	if (npixels != maxpixels) {			//	clear to EOBuffer
		if (gif.Interlace == FALSE) {
			bzero (gif.pic + npixels, (int)(maxpixels - npixels));
		}
	}
	
	OPaC_CloseFile (gif.fp);
	gif.fp = 0;
	
	OPGrailPixmapRef pixmap = OP_NEW_FROM_ZONE (zone, OPGrailPixmap);
	
	pixmap->Initialise ((unsigned)(0/*gif.BitsPerPixel*/),
						(OPCoord)(gif.Width), (OPCoord)(gif.Height));
	
	Card8* r = OP_ALLOC_ARRAY (Card8, gif.Width);
	Card8* g = OP_ALLOC_ARRAY (Card8, gif.Width);
	Card8* b = OP_ALLOC_ARRAY (Card8, gif.Width);
	
	a = 0;
	
	pixmap->BeginDraw (0, 0, (OPCoord)(gif.Width - 1), (OPCoord)(gif.Height - 1));
	
	for (j = 0; j < gif.Height; j++) {
		
		for (i = 0; i < gif.Width; i++) {
			
			r[i] = gif.r[gif.pic[a]];
			g[i] = gif.g[gif.pic[a]];
			b[i] = gif.b[gif.pic[a]];
			
			a++;
		}
		
		pixmap->SetRGBPixels (0, (OPCoord)(j), (OPCoord)(gif.Width), r, g, b);
	}
	
	pixmap->EndDraw ();
	
	OP_FREE (r);
	OP_FREE (g);
	OP_FREE (b);
	OP_FREE (gif.pic);
	OP_FREE (& gif);
	
	return pixmap;
}


/*
 *	Fetch the next code from the raster data stream. The
 *	codes can be any length from 3 to 12 bits, packed into
 *	8-bit bytes, so we have to maintain our location in the
 *	Raster array as a BIT Offset. We compute the byte Offset
 *	into the raster array by dividing this by 8, pick up
 *	three bytes, compute the bit Offset into our 24-bit chunk,
 *	shift to bring the desired code to the bottom, then mask
 *	it off and return it.
 */

static Int32
ReadCode (GIFloader& gif)
{
	OP_FUNCTION (ReadCode, 1);
	OP_ARG (1, GIFloader&, gif);
	
	Int32 RawCode;
	Int32 ByteOffset;
	
	ByteOffset = gif.BitOffset / 8;
	RawCode    = gif.Raster[ByteOffset] + ((Int32)(gif.Raster[ByteOffset + 1]) << 8);
	
	if (gif.CodeSize >= 8) {
		RawCode += ((Int32)(gif.Raster[ByteOffset + 2])) << 16;
	}
	
	RawCode      >>= gif.BitOffset % 8;
	gif.BitOffset += gif.CodeSize;
	
	return RawCode & gif.ReadMask;
}


static void
DoInterlace (GIFloader& gif, Card8 Index)
{
	OP_FUNCTION (DoInterlace, 2);
	OP_ARG (1, GIFloader&, gif);
	OP_ARG (2, Card8, Index);
	
	if (gif.oldYC != gif.YC) {
		gif.iPtr  = gif.pic + gif.YC * gif.Width;
		gif.oldYC = gif.YC;
	}
	
	if (gif.YC < gif.Height) *gif.iPtr++ = Index;
	
	//	Update the X-coordinate, and if it overflows, update the Y-coordinate.
	
	if (++gif.XC == gif.Width) {
		
		//	Deal with the interlace as described in the GIF spec. Put the
		//	decoded scan line out to the screen if we haven't gone past the
		//	bottom of it.
		
		gif.XC = 0;
		
		switch (gif.Pass) {
			
			case 0:
				gif.YC += 8;
				if (gif.YC >= gif.Height) { gif.Pass++; gif.YC = 4; }
				break;
			
			case 1:
				gif.YC += 8;
				if (gif.YC >= gif.Height) { gif.Pass++; gif.YC = 2; }
				break;
			
			case 2:
				gif.YC += 4;
				if (gif.YC >= gif.Height) { gif.Pass++; gif.YC = 1; }
				break;
			
			case 3:
				gif.YC += 2;
				break;
			
			default:
				break;
		}
	}
}



static OPGrailPixmap*
GifError (GIFloader& gif)
{
	OP_FUNCTION (GifError, 1);
	OP_ARG (1, GIFloader&, gif);
	
	if (gif.pic	!= 0) OP_FREE (gif.pic);
	if (gif.RawGIF != 0) OP_FREE (gif.RawGIF);
	if (gif.Raster != 0) OP_FREE (gif.Raster);
	if (gif.fp != 0) OPaC_CloseFile (gif.fp);
	
	OP_FREE (& gif);
	
	return 0;
}

